import collections
import os
import warnings

from dbm import dumb
from tempfile import mkstemp

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import gridspec, ticker
from scipy import stats

import varwg.time_series_analysis.seasonal_distributions as sd
import varwg.time_series_analysis.time_series as ts
from varwg import helpers as my
from varwg.core import base
from varwg.meteo import avrwind, meteox2y
from varwg.meteo.windrose import seasonal_windroses, windrose
from varwg.time_series_analysis import rain_stats

try:
    from varwg import config as conf
except ImportError:
    from varwg import config_template as conf

    conf_filepath = conf.__file__
    if conf_filepath.endswith(".pyc"):
        conf_filepath = conf_filepath[:-1]
    # warnings.warn('Could not import "config.py".\n' +
    #               'Edit "%s" and rename it to "config.py"' % conf_filepath)


def append_fa(fig_ax, figs=None, axs=None):
    if figs is None:
        figs = []
    if axs is None:
        axs = []
    assert len(figs) == len(axs)
    fig, ax = fig_ax
    return figs + [fig], axs + [ax]


def var_names_greek(var_names):
    return [conf.ygreek[var_name] for var_name in var_names]


class Plotting(base.Base):
    """To be used indirectly (by the VG class) or directly to plot
    meteorological data (met_files, pandas dataframe).

    Caching features will be disabled to not overwrite anything more useful.
    """

    def __init__(
        self,
        var_names,
        met_file=None,
        sum_interval=24,
        max_nans=12,
        plot=False,
        separator="\t",
        refit=None,
        detrend_vars=None,
        verbose=False,
        data_dir=None,
        cache_dir=None,
        seasonal_fitting=False,
        dump_data=False,
        non_rain=None,
        rain_method=None,
        fit_kwds=None,
        **met_kwds,
    ):
        # signal for Base to not do the seasonal fitting, which takes time
        # and the reason to initiate Plotting might be that you just plot
        # something quickly
        self.seasonal_fitting = seasonal_fitting
        super(Plotting, self).__init__(
            var_names,
            met_file=met_file,
            sum_interval=sum_interval,
            plot=plot,
            separator=separator,
            refit=refit,
            detrend_vars=detrend_vars,
            verbose=verbose,
            data_dir=data_dir,
            cache_dir=cache_dir,
            dump_data=dump_data,
            non_rain=non_rain,
            rain_method=rain_method,
            max_nans=max_nans,
            fit_kwds=fit_kwds,
            **met_kwds,
        )

    def plot_all(self, *args, **kwds):
        """Executes every method of Plotting that starts with "plot_".
        There are quite a number of those. You have been warned.
        """
        methods = {
            attr: getattr(self, attr)
            for attr in dir(Plotting)
            if (attr.startswith("plot_") and attr != "plot_all")
        }
        return_values = []
        for meth_name, meth in list(methods.items()):
            if self.verbose:
                print("Calling %s" % meth_name)
            try:
                fig_ax = meth(*args, **kwds)
                return_values += [fig_ax]
            except Exception as exc:
                warnings.warn("Exception while calling %s:\n%s" % (meth, exc))
                raise
        return return_values

    def fit_seasonal(self, refit=None, values=None, doys=None):
        """This is here, so that when initialized directly, seasonal
        distributions are fitted, but the existing fit to the data
        defined in the config file is not overwritten.
        """
        f_handle, self.seasonal_cache_file = mkstemp(suffix=".she")
        # need to establish the fact that the temp file is a data base
        db = dumb.open(self.seasonal_cache_file, "n")
        db.close()
        os.close(f_handle)
        self.data_trans, self.dist_sol = super(Plotting, self)._fit_seasonal(
            refit, values, doys
        )

    def __getattribute__(self, name):
        if name == "dist_sol" and not self.seasonal_fitting:
            # ok, we need the seasonal fitting, so provide it
            self.seasonal_fitting = True
            self.fit_seasonal(refit="all")
            return self.dist_sol
        return base.Base.__getattribute__(self, name)

    @property
    def ylabels(self):
        return [
            conf.units[var] if var in conf.units else "[?]"
            for var in self.var_names
        ]

    def plot_exceedance_daily(
        self, thresh=None, fig=None, axs=None, *args, **kwds
    ):
        if "R" not in self.var_names:
            warnings.warn("No R in var_names, no plot_exceedance_daily")
            return
        if thresh is None:
            thresh = conf.threshold
        rain_i = self.var_names.index("R")
        sim = self.sim_sea[rain_i]
        obs = self.data_raw[rain_i] / self.sum_interval[rain_i]
        fig, axs = rain_stats.plot_exceedance(
            obs,
            sim,
            kind="all",
            thresh=thresh,
            fig=fig,
            axs=axs,
            *args,
            **kwds,
        )
        title_str = "Precipitation exceedance"
        fig.suptitle(title_str)
        try:
            fig.canvas.set_window_title(
                f"{title_str} ({fig.canvas.manager.num})"
            )
        except AttributeError:
            pass
        return fig, axs

    def plot_exeedance_hourly(
        self, thresh=None, fig=None, axs=None, *args, **kwds
    ):
        if "R" not in self.var_names:
            warnings.warn("No R in var_names, no plot_exceedance_hourly")
            return
        if self.sim_sea is None:
            warnings.warn("Call simulate first.")
            return
        if self.sim_sea_dis is None:
            warnings.warn("No hourly simulation available")
            return
        if thresh is None:
            thresh = conf.threshold
        rain_i = self.var_names.index("R")
        sim = self.sim_sea_dis[rain_i]
        obs = self.met["R"]
        return rain_stats.plot_exceedance(
            obs,
            sim,
            kind="all",
            thresh=thresh,
            fig=None,
            axs=None,
            *args,
            **kwds,
        )

    def plot_hyd_year_sums(self, fig=None, ax=None, alpha=0.5, *args, **kwds):
        if "R" not in self.var_names:
            warnings.warn("No R in var_names, no plot_hyd_year_sums")
            return
        import xarray as xr

        if fig is None or ax is None:
            fig, ax = plt.subplots(nrows=1, ncols=1)
            first_fig = True
        else:
            first_fig = False
        obs_xr = (
            xr.DataArray(self.to_df("daily input"))
            .rename(dict(dim_1="variable"))
            .sel(variable="R")
        )
        obs = np.sort(rain_stats.hyd_year_sums(obs_xr, full_years=True))
        ax.step(obs, my.rel_ranks(len(obs)), where="mid", label="obs")

        if self.sim_sea is not None:
            sim_xr = (
                xr.DataArray(self.to_df("daily output"))
                .rename(dict(dim_1="variable"))
                .sel(variable="R")
            )
            sim = np.sort(rain_stats.hyd_year_sums(sim_xr, full_years=True))
            ax.step(
                sim,
                my.rel_ranks(len(sim)),
                where="mid",
                color="k",
                alpha=alpha,
                label="sim",
            )
            if first_fig:
                ax.legend(loc="best")

        if first_fig:
            ax.grid(True)
            ax.set_ylabel("cdf")
            ax.set_xlabel(f"R {conf.units['R']}")
            title_str = "Hydrological year sums (cdf)"
            ax.set_title(title_str)
            try:
                fig.canvas.set_window_title(
                    f"{title_str} ({fig.canvas.manager.num})"
                )
            except AttributeError:
                pass
        return fig, ax

    def plot_psd(self, var_names=None, hourly=False, *args, **kwds):
        """Plots power spectral density using matplotlib.pyplot.psd.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        hourly : boolean, optional
            Plot hourly and disaggregated values.
        *args, **kwds :
            Passed on to matplotlib.pyplot.psd
        """
        if var_names is None:
            var_names = self.var_names

        fig, axs = plt.subplots(len(var_names))
        if hourly:
            data_obs = [self.met[var_name] for var_name in var_names]
            data_sim = [
                self.sim_sea_dis[self.var_names.index(var_name)]
                for var_name in var_names
            ]
            # time discretization
            dt = (self.dis_times[1] - self.dis_times[0]).total_seconds()
            suptitle = "Power spectral density - hourly"
            name = "psd_hourly"
        else:
            data_obs = [
                self.data_raw[self.var_names.index(var_name)]
                for var_name in var_names
            ]
            data_sim = [
                self.sim_sea[self.var_names.index(var_name)]
                for var_name in var_names
            ]
            dt = (self.sim_times[1] - self.sim_times[0]).total_seconds()
            suptitle = "Power spectral density - daily"
            name = "psd_daily"

        for var_name in var_names:
            var_i = self.var_names.index(var_name)
            axs[var_i].psd(
                data_obs[var_i],
                Fs=1 / dt,
                scale_by_freq=True,
                label="observed",
                *args,
                **kwds,
            )
            axs[var_i].psd(
                data_sim[var_i],
                Fs=1 / dt,
                scale_by_freq=True,
                label="simulated",
                *args,
                **kwds,
            )
            axs[var_i].set_xscale("log")
            axs[var_i].set_ylabel("%s [dB/Hz]" % var_name)
        axs[0].legend(loc="best")
        fig.suptitle(suptitle)
        fig.name = name

        return fig, axs

    def plot_doy_scatter(
        self, var_names=None, opacity=0.4, hourly=False, **f_kwds
    ):
        """Plot variables over doys. Measurement and backtransformed simulated
        data will be plotted, if it is available.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        opacity : float in the range (0,1], optional
            Opacity of the scatter points.

        .. plot::

            met_vg = VG(("theta",))
            sim_times, sim = met_vg.simulate()
            met_vg.plot_doy_scatter()

        """
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, str):
            var_names = (var_names,)
        fig, axs = plt.subplots(nrows=len(var_names), **f_kwds)
        if len(var_names) == 1:
            axs = (axs,)
        for ax, var_name in zip(axs, var_names):
            var_ii = self.var_names.index(var_name)
            if hourly:
                ax.scatter(
                    varwg.times.datetime2doy(self.times_orig),
                    self.met[var_name],
                    marker="o",
                    edgecolors=(0, 0, 1, opacity),
                    facecolors=(0, 0, 0, 0),
                    label="measured",
                )
            else:
                ax.scatter(
                    self.data_doys,
                    self.data_raw[var_ii] / self.sum_interval[var_ii],
                    marker="o",
                    edgecolors=(0, 0, 1, opacity),
                    facecolors=(0, 0, 0, 0),
                    label="measured",
                )

            if self.sim_sea is not None:
                if hourly:
                    ax.scatter(
                        self.dis_doys,
                        self.sim_sea_dis[var_ii],
                        marker="o",
                        edgecolors=(1, 0, 0, opacity),
                        facecolors=(0, 0, 0, 0),
                        label="simulated backtransformed",
                    )
                else:
                    ax.scatter(
                        self.sim_doys,
                        self.sim_sea[var_ii],
                        marker="o",
                        edgecolors=(1, 0, 0, opacity),
                        facecolors=(0, 0, 0, 0),
                        label="simulated backtransformed",
                    )
            ax.set_xlabel("doy")
            ax.set_ylabel(
                "%s %s"
                % (
                    var_names_greek(self.var_names)[var_ii],
                    conf.units[var_name],
                )
            )
            ax.legend(loc="best")
            ax.grid(True)
            ax.set_xlim(0, 366)
            ax.set_title(conf.long_var_names[var_name])
        title_str = "Seasonal scatter"
        fig.suptitle(title_str)
        try:
            fig.canvas.set_window_title(
                f"{title_str} ({fig.canvas.manager.num})"
            )
        except AttributeError:
            pass
        return fig, ax

    def _meteogram(
        self,
        time,
        data,
        suptitle,
        var_names,
        fig=None,
        axs=None,
        plot_dewpoint=True,
        p_kwds=None,
        h_kwds=None,
        **f_kwds,
    ):
        if fig is None:
            fig = plt.figure(constrained_layout=True, **f_kwds)
        if var_names is None:
            var_names = self.var_names
        if isinstance(var_names, str):
            var_names = (var_names,)
        if p_kwds is None:
            p_kwds = {}
        if h_kwds is None:
            h_kwds = {}
        K = len(var_names)
        if axs is None:
            gs = gridspec.GridSpec(K, 2, figure=fig, width_ratios=[8, 1])
            axs = np.empty((K, 2), dtype=object)
            for i in range(K):
                if i == 0:
                    axs[0, 0] = plt.subplot(gs[0, 0])
                else:
                    axs[i, 0] = plt.subplot(gs[i, 0], sharex=axs[i - 1, 0])
                axs[i, 1] = plt.subplot(gs[i, 1], sharey=axs[i, 0])

        for var_name in var_names:
            var_i = var_names.index(var_name)
            plt.xticks(rotation=70)
            axs[var_i, 0].plot(time, data[var_i], label=var_name, **p_kwds)
            if (
                plot_dewpoint
                and var_name == "theta"
                and ("rh" in var_names or "abs_hum" in var_names)
            ):
                theta = data[var_i]
                if "rh" in var_names:
                    rh = data[var_names.index("rh")]
                else:
                    abs_hum = data[var_names.index("abs_hum")]
                    rh = meteox2y.abs_hum2rel(abs_hum, theta)
                dewpoint = meteox2y.dewpoint(theta, rh=rh)
                axs[var_i, 0].plot(
                    time,
                    dewpoint,
                    "--",
                    label="dewpoint",
                    color="k",
                    alpha=0.5,
                    **p_kwds,
                )
            if var_name == "abs_hum" and "theta" in var_names:
                theta = data[var_names.index("theta")]
                sat_vap_p = meteox2y.sat_vap_p(theta)
                axs[var_i, 0].plot(
                    time,
                    sat_vap_p,
                    "--",
                    label="sat_vap_p",
                    color="k",
                    alpha=0.5,
                    **p_kwds,
                )
            axs[var_i, 0].grid(True)
            axs[var_i, 0].set_ylabel(
                "%s %s" % (conf.ygreek[var_name], conf.units[var_name])
            )
            finite_mask = np.isfinite(data[var_i])
            try:
                axs[var_i, 1].hist(
                    data[var_i, finite_mask],
                    40,
                    density=True,
                    histtype="step",
                    orientation="horizontal",
                    **h_kwds,
                )
            except TypeError:
                pass
            # if var_i < (K - 1):
            #     axs[var_i, 0].set_xticklabels("")
            axs[var_i, 1].set_xticklabels("")
        plt.suptitle(suptitle)
        return fig, axs

    def _add_daily_bounds(self, axs, var_names, doys, times):
        keys = "u", "l", "uc", "lc"
        for var_i, ax in enumerate(axs[:, 0]):
            var_name = var_names[var_i]
            if var_name == "R":
                continue
            if var_name not in conf.par_known:
                continue
            pars = conf.par_known[var_name]
            if pars is None:
                continue
            for key in keys:
                if key not in pars:
                    continue
                par_values = pars[key](doys) / self.sum_interval_dict[var_name]
                ax.plot(times, par_values, label=key)
            ax.legend(loc="best")
        return axs

    def plot_meteogram_daily(
        self,
        var_names=None,
        figs=None,
        axss=None,
        figsize=None,
        plot_sim_sea=True,
        obs=None,
        sim=None,
        plot_daily_bounds=True,
        p_kwds=None,
        h_kwds=None,
        **f_kwds,
    ):
        if var_names is None:
            var_names = self.var_names
        if figs is None:
            figs = 2 * [None]
        elif isinstance(figs, mpl.figure.Figure):
            figs = (figs,)
        if axss is None:
            axss = 2 * [None]
        if figsize is None:
            figsize = (8, len(var_names))
        if figsize is not None:
            f_kwds["figsize"] = figsize
        # elif isinstance(axss, np.ndarray):
        #     axss = axss,
        if self.ex_in is None and obs is None:
            obs = self.data_raw / self.sum_interval
        elif obs is None:
            obs = np.vstack((self.data_raw / self.sum_interval, self.ex_in))
            if "R" in var_names:
                rain_i = var_names.index("R")
                obs[rain_i] *= self.sum_interval[rain_i]
            var_names = list(var_names) + ["external"]

        fig, axs = self._meteogram(
            self.times,
            obs,
            "Measured daily",
            var_names,
            fig=figs[0],
            axs=axss[0],
            p_kwds=p_kwds,
            h_kwds=h_kwds,
            **f_kwds,
        )
        fig.name = "meteogram_measured_daily"
        if plot_daily_bounds:
            axs = self._add_daily_bounds(
                axs, var_names, self.data_doys, self.times
            )

        if plot_sim_sea and self.sim_sea is not None:
            if self.ex_out is None and sim is None:
                sim = self.sim_sea
            elif sim is None:
                sim = np.vstack((self.sim_sea, self.ex_out))
                var_names = list(var_names) + ["external"]
            fig_, axs_ = self._meteogram(
                self.sim_times,
                sim,
                "Simulated daily",
                var_names,
                fig=figs[1],
                axs=axss[1],
                p_kwds=p_kwds,
                h_kwds=h_kwds,
                **f_kwds,
            )
            fig_.name = "meteogram_sim_daily"
            if plot_daily_bounds:
                axs_ = self._add_daily_bounds(
                    axs_, var_names, self.sim_doys, self.sim_times
                )
            fig, axs = np.array([fig, fig_]), np.array([axs, axs_])
        return fig, axs

    def plot_meteogram_hourly(
        self,
        var_names=None,
        figs=None,
        axss=None,
        plot_sim_sea=True,
        obs=None,
        sim=None,
        combine=False,
        **f_kwds,
    ):
        """All variables over time in subplots.
        (Not as nice as the meteogram from varwg.meteo)
        """
        if figs is None:
            figs = 2 * [None]
        if axss is None:
            axss = 2 * [None]
        if obs is None:
            if var_names is None:
                var_names = self.var_names
            obs = base.met_as_array(self.met, var_names=var_names)
        fig, axs = self._meteogram(
            self.times_orig,
            obs,
            "Measured hourly",
            var_names,
            fig=figs[0],
            axs=axss[0],
        )
        fig.name = "meteogram_measured_houry"

        if plot_sim_sea and self.sim_sea_dis is not None:
            if self.ex_out is None and sim is None:
                sim = self.sim_sea_dis
                var_names = self.var_names
            elif sim is None:
                sim = np.vstack((self.sim_sea_dis, self.ex_out))
                var_names = list(self.var_names) + ["external"]
            if combine:
                fig, axs = self._meteogram(
                    self.dis_times,
                    sim,
                    "Simulated daily",
                    var_names,
                    fig=fig,
                    axs=axs,
                )
            else:
                fig_, axs_ = self._meteogram(
                    self.dis_times,
                    sim,
                    "Simulated daily",
                    var_names,
                    fig=figs[1],
                    axs=axss[1],
                )
                fig_.name = "meteogram_sim_daily"
                fig, axs = np.array([fig, fig_]), np.array([axs, axs_])

        # if self.sim_sea_dis is not None:
        #     fig_, axs_ = self._meteogram(self.dis_times, self.sim_sea_dis,
        #                                   "Simulated hourly", var_names,
        #                                   fig=figs[1], axs=axss[1],
        #                                   station_name=self.station_name)
        #     fig_.name = "meteogram_sim_hourly"
        #     fig, axs = np.array([fig, fig_]), np.array([axs, axs_])
        return fig, axs

    def plot_meteogram_trans(
        self, var_names=None, figs=None, axss=None, **f_kwds
    ):
        if figs is None:
            figs = 2 * [None]
        elif isinstance(figs, mpl.figure.Figure):
            figs = (figs,)
        if axss is None:
            axss = 2 * [None]
        elif isinstance(axss, np.ndarray):
            axss = (axss,)
        fig, axs = self._meteogram(
            self.times,
            self.data_trans,
            "Observed transformed",
            var_names,
            fig=figs[0],
            axs=axss[0],
            plot_dewpoint=False,
        )
        fig.name = "meteogram_measured_daily_trans"
        if self.sim is not None:
            fig_, axs_ = self._meteogram(
                self.sim_times,
                self.sim,
                "Simulated std-normal",
                var_names,
                fig=figs[1],
                axs=axss[1],
                plot_dewpoint=False,
            )
            fig_.name = "meteogram_sim_trans"
            fig, axs = np.array([fig, fig_]), np.array([axs, axs_])
        return fig, axs

    def plot_doy_scatter_residuals(self, var_names=None, opacity=0.4):
        if var_names is None:
            var_names = self.var_names
        # when given a string, this does not separate the letters, as tuple()
        # or list does
        var_names = np.atleast_1d(var_names)
        figs = []
        axs = []
        for var_name in var_names:
            var_ii = self.var_names.index(var_name)
            fig, ax = plt.subplots(nrows=1, ncols=1)
            figs.append(fig)
            axs.append(ax)
            ax.scatter(
                self.data_doys,
                self.data_trans[var_ii],
                marker="o",
                edgecolors=(0, 0, 1, opacity),
                facecolors=(0, 0, 0, 0),
                label="transformed",
            )

            ax.scatter(
                self.data_doys,
                self.residuals[var_ii],
                marker="o",
                edgecolors=(1, 0, 0, opacity),
                facecolors=(0, 0, 0, 0),
                label="residuals",
            )
            ax.set_xlabel("doy")
            ax.set_ylabel(
                "%s %s"
                % (
                    var_names_greek(self.var_names)[var_ii],
                    conf.units[var_name],
                )
            )
            ax.legend()
            ax.grid()
            title_str = (
                "Seasonal scatter of residuals %s"
                % conf.long_var_names[var_name]
            )
            ax.set_title(title_str)
            try:
                fig.canvas.set_window_title(
                    f"{title_str} ({fig.canvas.manager.num})"
                )
            except AttributeError:
                pass
        return figs, axs

    def plot_spaces(self, var_names=None, opacity=0.4):
        if var_names is None:
            var_names = self.var_names

        figs = []
        axs = []
        for var_name1 in var_names:
            var_ii = self.var_names.index(var_name1)
            for var_jj in range(var_ii + 1, len(var_names)):
                var_name2 = var_names[var_jj]
                fig, ax = plt.subplots(
                    nrows=1, ncols=1, subplot_kw=dict(aspect="equal")
                )
                figs.append(fig)
                axs.append(ax)
                ax.scatter(
                    my.rel_ranks(self.data_trans[var_ii]),
                    my.rel_ranks(self.data_trans[var_jj]),
                    marker="o",
                    edgecolors=(0, 0, 1, opacity),
                    facecolors=(0, 0, 0, 0),
                    label="transformed data",
                )
                ax.scatter(
                    my.rel_ranks(self.sim[var_ii]),
                    my.rel_ranks(self.sim[var_jj]),
                    marker="o",
                    edgecolors=(1, 0, 0, opacity),
                    facecolors=(0, 0, 0, 0),
                    label="simulated",
                )
                ax.legend()
                ax.set_xlabel(var_name1)
                ax.set_ylabel(var_name2)
                ax.grid()
                title_str = "Spaces %s over %s" % (
                    conf.long_var_names[var_name2],
                    conf.long_var_names[var_name1],
                )
                ax.set_title(title_str)
                try:
                    fig.canvas.set_window_title(
                        f"{title_str} ({fig.canvas.manager.num})"
                    )
                except AttributeError:
                    pass
        return figs, axs

    def plot_diff_spaces(self, var_names=None, diff=1, opacity=0.4):
        if var_names is None:
            var_names = self.var_names

        figs = []
        axs = []
        for var_name in var_names:
            var_ii = self.var_names.index(var_name)
            fig, ax = plt.subplots(
                nrows=1, ncols=1, subplot_kw=dict(aspect="equal")
            )
            figs.append(fig)
            axs.append(ax)
            ax.scatter(
                my.rel_ranks(self.data_trans[var_ii, diff:]),
                my.rel_ranks(self.data_trans[var_ii, :-diff]),
                marker="o",
                edgecolors=(0, 0, 1, opacity),
                facecolors=(0, 0, 0, 0),
                label="transformed data",
            )
            ax.scatter(
                my.rel_ranks(self.sim[var_ii, diff:]),
                my.rel_ranks(self.sim[var_ii, :-diff]),
                marker="o",
                edgecolors=(1, 0, 0, opacity),
                facecolors=(0, 0, 0, 0),
                label="simulated",
            )
            ax.legend()
            ax.set_xlabel("t")
            ax.set_ylabel("t - %d" % diff)
            ax.grid(True)
            title_str = "Spaces %d-lagged %s" % (
                diff,
                conf.long_var_names[var_name],
            )
            ax.set_title(title_str)
            try:
                fig.canvas.set_window_title(
                    f"{title_str} ({fig.canvas.manager.num})"
                )
            except AttributeError:
                pass
        return figs, axs

    def plot_hourly_fit(self, var_names=None, *args, **kwds):
        """Plots hourly scatter with fitted densities and a histogram of the
        transformed quantiles.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        plot_quantiles : boolean, optional
            Plot histograms of the quantiles?
        plot_fourier : boolean, optional
            Plot the fourier fit to the seasonally changing distribution
            parameters.
        plot_monthly : boolean, optional
            Plot histograms showing how good the fit is on monthly separated
            data.
        opacity : float, optional
            Opacity of scatter.
        n_bins : int or sequence, optional
            Passed on to np.histogram when plotting histograms of quantiles.
        kde : boolean, optional
            Plot a kernel density estimation on the histograms of quantiles.
        s_kwds : None or dict, optional
            Keyword arguments passed on to plt.scatter.
        """
        return self._plot_seasonal_fit(
            var_names=None, hourly=True, *args, **kwds
        )

    def plot_daily_fit(self, var_names=None, *args, **kwds):
        """Plots daily scatter with fitted densities and a histogram of the
        transformed quantiles.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        plot_quantiles : boolean, optional
            Plot histograms of the quantiles?
        plot_fourier : boolean, optional
            Plot the fourier fit to the seasonally changing distribution
            parameters.
        plot_monthly : boolean, optional
            Plot histograms showing how good the fit is on monthly separated
            data.
        opacity : float, optional
            Opacity of scatter.
        n_bins : int or sequence, optional
            Passed on to np.histogram when plotting histograms of quantiles.
        kde : boolean, optional
            Plot a kernel density estimation on the histograms of quantiles.
        s_kwds : None or dict, optional
            Keyword arguments passed on to plt.scatter.
        """
        return self._plot_seasonal_fit(
            var_names=var_names, hourly=False, *args, **kwds
        )

    def _plot_seasonal_fit(
        self,
        var_names=None,
        plot_quantiles=False,
        plot_fourier=False,
        plot_monthly=False,
        plot_seasonality=False,
        hourly=False,
        opacity=0.25,
        n_bins=15,
        kde=True,
        figs=None,
        axss=None,
        s_kwds=None,
        *args,
        **kwds,
    ):
        """Plots seasonal scatter with fitted densities and a histogram of the
        transformed quantiles.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        plot_quantiles : boolean, optional
            Plot histograms of the quantiles?
        plot_fourier : boolean, optional
            Plot the fourier fit to the seasonally changing distribution
            parameters.
        plot_monthly : boolean, optional
            Plot histograms showing how good the fit is on monthly separated
            data.
        opacity : float, optional
            Opacity of scatter.
        n_bins : int or sequence, optional
            Passed on to np.histogram when plotting histograms of quantiles.
        kde : boolean, optional
            Plot a kernel density estimation on the histograms of quantiles.
        s_kwds : None or dict, optional
            Keyword arguments passed on to plt.scatter.
        """
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, str):
            var_names = (var_names,)

        if figs is None:
            figs = []
        if axss is None:
            axss = []
        for var_name in var_names:
            var_ii = self.var_names.index(var_name)
            if hourly:
                try:
                    dist, solution = self.dist_sol_hourly(var_name)
                except KeyError:
                    warnings.warn("No hourly fit for %s available" % var_name)
                    continue
                n_sumup = 1
                values = self.met[var_name]
                doys = self.data_doys_raw
                discr_str = "hourly"
            else:
                dist, solution = self.dist_sol[var_name]
                n_sumup = self.sum_interval.ravel()[var_ii]
                values = self.data_raw[var_ii]
                doys = self.data_doys
                discr_str = "daily"

            if var_name in conf.long_var_names:
                long_var_name = conf.long_var_names[var_name]
            else:
                long_var_name = var_name

            # scatter pdf
            fig, ax = dist.scatter_pdf(
                solution,
                title=long_var_name,
                n_sumup=n_sumup,
                opacity=opacity,
                s_kwds=s_kwds,
                *args,
                **kwds,
            )
            if self.sim_sea is not None:
                # data_sim = (self.sim_sea[var_ii] *
                #             self.sum_interval_dict[var_name])
                data_sim = self.sim_sea[var_ii]
                ax.scatter(
                    self.sim_doys,
                    data_sim,
                    marker="x",
                    facecolor=(0, 0, 0, opacity),
                )
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(
                    min(ymin, data_sim.min()), max(ymax, data_sim.max())
                )
            fig.name = "scatter_pdf_%s_%s" % (var_name, discr_str)
            ax.set_ylabel(
                conf.units[var_name] if var_name in conf.units else "[?]"
            )
            title_str = "Scatter pdf %s (%d) %s" % (
                var_name,
                fig.canvas.manager.num,
                discr_str,
            )
            ax.set_title(title_str)
            try:
                fig.canvas.set_window_title(title_str)
            except AttributeError:
                pass
            figs += [fig]
            axss += [ax]

            if plot_quantiles:
                fig = plt.figure(*args, **kwds)
                var_ii = self.var_names.index(var_name)
                quantiles = dist.cdf(solution, x=values, doys=doys)
                my.hist(
                    quantiles[np.isfinite(quantiles)], n_bins, fig=fig, kde=kde
                )
                title_str = "Quantiles %s (%d) %s" % (
                    var_name,
                    fig.canvas.manager.num,
                    discr_str,
                )
                plt.title(title_str)
                try:
                    fig.canvas.set_window_title(title_str)
                except AttributeError:
                    pass
                fig.name = "scatter_pdf_quantiles_%s_%s" % (
                    var_name,
                    discr_str,
                )
                figs += [fig]

            suptitle = "%s %s" % (var_name, discr_str)
            if plot_fourier:
                if isinstance(dist, sd.SeasonalDist):
                    fig, _ = dist.plot_fourier_fit()
                    fig.suptitle(suptitle)
                    try:
                        fig.canvas.set_window_title(suptitle)
                    except AttributeError:
                        pass
                    fig.name = "scatter_pdf_fft_%s_%s" % (var_name, discr_str)
                    figs += [fig]
            if plot_monthly:
                if isinstance(dist, sd.SlidingDist):
                    fig, axs = dist.plot_monthly_fit()
                    try:
                        for f in fig:
                            f.suptitle(suptitle)
                            try:
                                fig.canvas.set_window_title(suptitle)
                            except AttributeError:
                                pass
                    except TypeError:
                        fig.suptitle(suptitle)
                        try:
                            fig.canvas.set_window_title(suptitle)
                        except AttributeError:
                            pass
                    if isinstance(fig, mpl.figure.Figure):
                        fig = [fig]
                    if isinstance(axs, np.ndarray):
                        axs = axs.tolist()
                    figs += fig
                    axss += axs
            if plot_seasonality:
                if isinstance(dist, sd.SlidingDist):
                    fig, axs = dist.plot_seasonality_fit()
                    try:
                        for f in fig:
                            f.suptitle(suptitle)
                            try:
                                fig.canvas.set_window_title(suptitle)
                            except AttributeError:
                                pass

                    except TypeError:
                        fig.suptitle(suptitle)
                        try:
                            fig.canvas.set_window_title(suptitle)
                        except AttributeError:
                            pass
                    if isinstance(fig, mpl.figure.Figure):
                        fig = [fig]
                    if isinstance(axs, np.ndarray):
                        axs = axs.tolist()
                    figs += fig
                    axss += axs
        return figs, axss

    def plot_qq(
        self,
        var_names=None,
        figsize=None,
        lines=True,
        trans=False,
        color="gray",
        fig=None,
        axs=None,
        obs=None,
        *args,
        **kwds,
    ):
        if var_names is None:
            var_names = self.var_names
        if isinstance(var_names, str):
            var_names = (var_names,)
        if obs is None:
            obs_all = self.data_raw
        else:
            obs_all = obs
        n_axes = len(var_names)
        n_cols = int(np.ceil(n_axes**0.5))
        n_rows = int(np.ceil(float(n_axes) / n_cols))
        if fig is None and axs is None:
            fig, axs = plt.subplots(
                n_rows, n_cols, subplot_kw={"aspect": 1}, figsize=figsize
            )
        alphas = np.linspace(0, 1, 200)
        try:
            axs = axs.ravel()
        except AttributeError:
            axs = np.array([axs])
        for ax_i, var_name in enumerate(var_names):
            var_ii = self.var_names.index(var_name)
            finite_obs = np.isfinite(obs_all[var_ii])
            if trans:
                obs = np.sort(np.copy(self.data_trans[var_ii, finite_obs]))
                # sim = np.sort(np.copy(self.sim[var_ii, finite_obs]))
                sim = np.sort(np.copy(self.sim[var_ii]))
            else:
                obs = np.sort(np.copy(obs_all[var_ii, finite_obs]))
                obs /= self.sum_interval[var_ii]
                # sim = np.sort(np.copy(self.sim_sea[var_ii, finite_obs]))
                sim = np.sort(np.copy(self.sim_sea[var_ii]))
            obs = np.quantile(obs, alphas)
            sim = np.quantile(sim, alphas)
            ax = axs[ax_i]
            if lines:
                ax.plot(sim, obs, color=color)
            else:
                ax.scatter(sim, obs, marker="x", color=color)
            global_min = min(obs[0], sim[0])
            global_max = max(np.nanmax(obs), sim[-1])

            if (
                self.sim_sea_dis is not None
                and self.var_names_dis is not None
                and var_name in self.var_names_dis
            ):
                obs_dis_all = np.array(self.met[var_name])[
                    : self.sim_sea_dis.shape[1]
                ]
                finite_obs = np.isfinite(obs_dis_all)
                var_i_dis = self.var_names_dis.index(var_name)
                obs_dis = np.sort(np.copy(obs_dis_all[finite_obs]))
                sim_dis = np.squeeze(self.sim_sea_dis[var_i_dis, finite_obs])
                sim_dis = np.sort(np.copy(sim_dis))
                if lines:
                    ax.plot(sim_dis, obs_dis, color="black")
                else:
                    ax.scatter(sim_dis, obs_dis, marker="x", color="black")
                global_min = min(global_min, obs_dis[0], sim_dis[0])
                global_max = max(
                    global_max, np.nanmax(obs_dis), np.nanmax(sim_dis)
                )

            ax.plot([global_min, global_max], [global_min, global_max], "k--")
            ax.grid()
            ax.set_ylabel("observed")
            if ax_i < (n_axes - n_cols):
                ax.set_xticklabels([])
            else:
                ax.set_xlabel("simulated")
            ax.set_title(
                "%s %s" % (conf.ygreek[var_name], conf.units[var_name])
            )
        # delete axs that we did not use
        if len(axs) > n_axes:
            for ax in axs[n_axes:]:
                ax.set_axis_off()
            plt.draw()
        title_str = "QQ-plots"
        fig.suptitle(title_str)

        try:
            fig.canvas.set_window_title(title_str)
        except AttributeError:
            pass
        fig.name = "qq"
        return fig, axs

    def plot_seasonal_corrs(
        self, var_names=None, trans=False, figsize=None, fig=None, axs=None
    ):
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, str):
            var_names = (var_names,)
        if fig is None and axs is None:
            fig, axs = plt.subplots(
                nrows=len(var_names),
                ncols=1,
                sharex=True,
                constrained_layout=True,
                figsize=figsize,
            )
        if trans:
            obs_df = self.to_df("daily input trans", var_names=var_names)
            sim_df = self.to_df("daily output trans", var_names=var_names)
        else:
            obs_df = self.to_df("daily input", var_names=var_names)
            sim_df = self.to_df("daily output", var_names=var_names)
        for ax, var_name in zip(axs, var_names):
            corrs_obs = np.empty((12, self.K - 1), dtype=float)
            var_i = var_names.index(var_name)
            var_other_ii = [
                idx for idx in range(len(var_names)) if idx != var_i
            ]
            for month_i, group in obs_df.groupby(obs_df.index.month):
                corrs = np.corrcoef(group.T)
                corrs_obs[month_i - 1] = corrs[var_i, var_other_ii]
            # simulated time series can have a different length
            corrs_sim = np.empty((12, self.K - 1), dtype=float)
            for month_i, group in sim_df.groupby(sim_df.index.month):
                corrs = np.corrcoef(group.T)
                corrs_sim[month_i - 1] = corrs[var_i, var_other_ii]
            var_names_other = [var_names[idx] for idx in var_other_ii]
            for corr_obs in corrs_obs.T:
                ax.plot(corr_obs, "--")
            ax.set_prop_cycle(None)
            for corr_sim, var_name_other in zip(corrs_sim.T, var_names_other):
                ax.plot(corr_sim, label=var_name_other)
            ax.set_title(var_name)
            ax.legend(loc="best")
            ax.grid(True, zorder=0)
        fig.suptitle(f"Monthly correlations {'trans' if trans else ''}")
        return fig, axs

    def plot_windrose(self, figsize=None, seasonal=False, *args, **kwds):
        try:
            ui, vi = self.var_names.index("u"), self.var_names.index("v")
        except ValueError:
            print("Can only plot wind roses if wind speed is simulated")
            return

        u = self.data_raw[ui] / self.sum_interval[ui]
        v = self.data_raw[vi] / self.sum_interval[vi]
        wind_dirs = avrwind.component2angle(u, v)[0]

        if figsize is not None:
            fig = plt.figure(figsize=figsize)
        else:
            fig = plt.figure()

        def rose_func_gen(times_):
            if seasonal:
                return lambda *args, **kwds: seasonal_windroses(
                    times_, *args, **kwds
                )
            else:
                return lambda *args, **kwds: windrose(*args, **kwds)

        rose_func = rose_func_gen(self.times)
        figs = [rose_func(wind_dirs, fig=fig, *args, **kwds)]
        figs[0].suptitle("measured")

        if self.sim_sea is not None:
            u, v = self.sim_sea[ui], self.sim_sea[vi]
            wind_dirs = avrwind.component2angle(u, v)[0]
            if figsize is not None:
                fig = plt.figure(figsize=figsize)
            else:
                fig = plt.figure()
            figs += [rose_func(wind_dirs, fig=fig, *args, **kwds)]
            figs[-1].suptitle("simulated")

        if self.sim_sea_dis is not None:
            rose_func = rose_func_gen(self.times_orig)
            u, v = self.met["u"], self.met["v"]
            wind_dirs = avrwind.component2angle(u, v)[0]
            if figsize is not None:
                fig = plt.figure(figsize=figsize)
            else:
                fig = plt.figure()
            figs += [rose_func(wind_dirs, fig=fig, *args, **kwds)]
            figs[-1].suptitle("measured hourly")

            rose_func = rose_func_gen(self.dis_times)
            ui_dis = self.var_names_dis.index("u")
            vi_dis = self.var_names_dis.index("v")
            u, v = self.sim_sea_dis[ui_dis], self.sim_sea_dis[vi_dis]
            wind_dirs = avrwind.component2angle(u, v)[0]
            if figsize is not None:
                fig = plt.figure(figsize=figsize)
            else:
                fig = plt.figure()
            figs += [rose_func(wind_dirs, fig=fig, *args, **kwds)]
            figs[-1].suptitle("simulated disaggregated")

        return np.squeeze(figs)

    def plot_corr(self, ygreek=None, hourly=False, trans=False, *args, **kwds):
        if ygreek is None:
            # get rid of the incident for this block
            ygreek = collections.defaultdict(lambda: r"-")
            ygreek.update(
                {
                    "R": r"$R$",
                    "theta": r"$\theta$",
                    "rh": r"$\phi$",
                    "Qsw": r"$Q_{sw}$",
                    "ILWR": r"$Q_{lw(i.)}$",
                    "u": r"$u$",
                    "v": r"$v$",
                }
            )

        if self.ex_in is None:
            data = self.data_raw
            var_names = self.var_names
        else:
            data = np.vstack((self.data_raw, self.ex_in))
            var_names = list(self.var_names) + ["external"]

        def my_greek(var_names):
            return [ygreek[var_name] for var_name in var_names]

        greek_short = my_greek(var_names)
        figs, axs = append_fa(
            ts.corr_img(data, 0, "Measured daily", greek_short, *args, **kwds)
        )
        figs[0].name = "corr_measured_daily"

        if trans:
            figs, axs = append_fa(
                ts.corr_img(
                    self.data_trans,
                    0,
                    "Measured daily transformed",
                    greek_short,
                    *args,
                    **kwds,
                ),
                figs,
                axs,
            )
            figs[-1].name = "corr_measured_daily_trans"

        if self.sim_sea is not None:
            if self.ex_out is None:
                data = self.sim_sea
                var_names = self.var_names
            else:
                data = np.vstack((self.sim_sea, self.ex_out))
                var_names = list(self.var_names) + ["external"]
            greek_short = my_greek(var_names)
            figs, axs = append_fa(
                ts.corr_img(
                    data, 0, "Simulated daily", greek_short, *args, **kwds
                ),
                figs,
                axs,
            )
            figs[-1].name = "corr_sim_daily"
            if trans:
                figs, axs = append_fa(
                    ts.corr_img(
                        self.sim,
                        0,
                        "Simulated daily transformed",
                        greek_short,
                        *args,
                        **kwds,
                    ),
                    figs,
                    axs,
                )
                figs[-1].name = "corr_sim_daily_trans"

        if hourly:
            met_array = base.met_as_array(self.met, var_names=self.var_names)
            figs, axs = append_fa(
                ts.corr_img(
                    met_array, 0, "Measured hourly", greek_short, *args, **kwds
                ),
                figs,
                axs,
            )
            figs[-1].name = "corr_measured_hourly"

            if self.sim_sea_dis is not None:
                figs, axs = append_fa(
                    ts.corr_img(
                        self.sim_sea_dis,
                        0,
                        "Simulated hourly",
                        greek_short,
                        *args,
                        **kwds,
                    ),
                    figs,
                    axs,
                )
                figs[-1].name = "corr_sim_hourly"
        return figs, axs

    def plot_VAR_par(self, **kwds):
        """Plots the parameters of the VAR-Process."""
        if self.AM is None:
            print("Call fit first")
            return
        return ts.matr_img(
            np.asarray(self.AM),
            "AM p=%d q=%d" % (self.p, self.q),
            None,
            self.var_names,
            **kwds,
        )

    def plot_autocorr(
        self, maxlag=7, title=None, var_names=None, *args, **kwds
    ):
        """Plots autocorrelation of transformed and simulated data (if the
        latter is available).

        Parameters
        ----------
        maxlag : int, optional
            Maximum number of lags to plot
        """
        data = self.data_trans
        figs = []
        if self.sim is not None:
            data = [data, self.sim]
        if title is None:
            title_ = "Autocorrelation of standard-normal distributed variables"
        if var_names is None:
            var_names = var_names_greek(self.var_names)
        figs, axs = append_fa(
            ts.plot_auto_corr(data, maxlag, title_, var_names, *args, **kwds)
        )
        data = self.data_raw
        if self.sim_sea is not None:
            data = [data, self.sim_sea]
        if title is None:
            title_ = "Autocorrelation of untransformed variables"
        figs, axs = append_fa(
            ts.plot_auto_corr(data, maxlag, title_, var_names, *args, **kwds),
            figs,
            axs,
        )
        if self.residuals is not None:
            figs, axs = append_fa(
                ts.plot_auto_corr(
                    self.residuals,
                    maxlag,
                    "Autocorrelation of residuals.",
                    var_names,
                    **kwds,
                ),
                figs,
                axs,
            )
            # otherwise we probably have been given two arrays.  then
            # things get more complicated, because we would distinguish
            # between two possibly different sample sizes (data.shape[1])
            if isinstance(data, np.ndarray):
                # bounds for test on the whiteness of the residuals (see page 160)
                bound = 2.0 / np.sqrt(data.shape[1]) * np.ones(maxlag)
                plt.plot(list(range(maxlag)), bound, "--", color="grey")
                plt.plot(list(range(maxlag)), -bound, "--", color="grey")

            if self.ut is None:
                squared_res = self.residuals**2
            else:
                squared_res = [self.residuals**2, self.ut**2]
            figs, axs = append_fa(
                ts.plot_auto_corr(
                    squared_res,
                    maxlag,
                    "Autocorrelation of " "squared residuals.",
                    var_names,
                    **kwds,
                ),
                figs,
                axs,
            )
        return figs, axs

    def plot_cross_corr(self, max_lags=7, figsize=None):
        """Plot cross correlations."""
        kwds = dict(var_names=self.var_names, max_lags=max_lags)
        fig, axs = ts.plot_cross_corr(
            self.data_raw / self.sum_interval, figsize=figsize, **kwds
        )
        if self.sim_sea is not None:
            fig, axs = ts.plot_cross_corr(
                self.sim_sea, linestyle="--", fig=fig, axs=axs, **kwds
            )
        suptitle = "Daily"
        plt.suptitle(suptitle)

        if self.sim_sea_dis is not None:
            met_array = base.met_as_array(self.met, var_names=self.var_names)
            kwds["max_lags"] *= max(self.sum_interval)
            fig1, axs1 = ts.plot_cross_corr(met_array, figsize=figsize, **kwds)
            fig1, axs1 = ts.plot_cross_corr(
                self.sim_sea_dis, linestyle="--", fig=fig1, axs=axs1, **kwds
            )
            suptitle = "Hourly"
            plt.suptitle(suptitle)
            fig = [fig, fig1]
            axs = [axs, axs1]

        return fig, axs

    def plot_scaling(
        self,
        agg_funcs=(np.mean, np.std, stats.skew, stats.kurtosis),
        max_agg_len=2 * 365,
        normalize=True,
        **fig_kw,
    ):
        """Plot mean/std/skew over aggregation length."""
        kwds = dict(
            var_names=self.var_names,
            agg_funcs=agg_funcs,
            max_agg_len=max_agg_len,
            **fig_kw,
        )
        color = kwds.pop("color", "blue")
        fig = kwds.pop("figs", None)
        axs = kwds.pop("axss", None)
        if fig is None and axs is None:
            fig, axs = ts.plot_scaling(
                self.data_raw / self.sum_interval, color="black", **kwds
            )

        if self.sim_sea is not None:
            sim = self.sim_sea
            if normalize:
                # for the sake of argument: make the means equal
                obs_means = np.nanmean(
                    self.data_raw / self.sum_interval, axis=1
                )
                sim_means = np.nanmean(self.sim_sea, axis=1)
                sim += (obs_means - sim_means)[:, None]
            fig, axs = ts.plot_scaling(
                sim, color=color, fig=fig, axs=axs, **kwds
            )
        return fig, axs

    def plot_monthly_hists(
        self, var_names=None, bins=20, figs=None, axss=None
    ):
        """Plot histograms grouped by month and variable."""
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, str):
            var_names = (var_names,)

        if figs is None:
            figs = []
            reuse_figs = False
        else:
            reuse_figs = True
        data = self.data_raw / self.sum_interval
        month_ii = [
            varwg.times.time_part(self.times, "%m") == month
            for month in range(1, 13)
        ]
        if self.sim_sea is not None:
            sim_month_ii = [
                varwg.times.time_part(self.sim_times, "%m") == month
                for month in range(1, 13)
            ]
        for var_name in var_names:
            var_i = self.var_names.index(var_name)
            if reuse_figs:
                fig, axs = figs[var_i], axss[var_i]
            else:
                fig, axs = plt.subplots(3, 4, sharey=True, sharex=True)
                axs = axs.ravel()
            for month in range(12):
                dat = data[var_i, month_ii[month]]
                dat = dat[np.isfinite(dat)]
                label = "obs"
                if self.sim_sea is not None:
                    label = [label, "sim"]
                    dat = [dat, self.sim_sea[var_i, sim_month_ii[month]]]
                axs[month].hist(
                    dat, bins, density=True, histtype="step", label=label
                )
                if self.sim_sea is None:
                    dat = (dat,)
                ax_cdf = axs[month].twinx()
                for values in dat:
                    values = values[np.isfinite(values)]
                    ranks = (np.arange(len(values)) - 0.5) / len(values)
                    ax_cdf.plot(np.sort(values), ranks)
                    ax_cdf.set_yticklabels([])
                ax_cdf.set_ylim(0, 1)
                axs[month].set_title(month + 1)
            if self.sim_sea is not None:
                axs[0].legend(loc="best")

            suptitle = var_name
            fig.suptitle(suptitle)
            fig.tight_layout(rect=(0, 0, 1, 0.95))
            if not reuse_figs:
                figs += [fig]
        if len(figs) == 1 and not reuse_figs:
            figs = figs[0]
        for ax in axs:
            ax.grid(True)
        return figs, axs

    def plot_monthly_hists_hourly(self, var_names=None, bins=20):
        """Plot histograms grouped by month and variable."""
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, str):
            var_names = (var_names,)

        figs = []
        data = base.met_as_array(self.met, var_names=var_names)
        month_ii = [
            varwg.times.time_part(self.times_orig, "%m") == month
            for month in range(1, 13)
        ]
        if self.sim_sea_dis is not None:
            sim_month_ii = [
                varwg.times.time_part(self.dis_times, "%m") == month
                for month in range(1, 13)
            ]
        for var_name in var_names:
            var_i = self.var_names.index(var_name)
            fig, axs = plt.subplots(3, 4, sharey=True, sharex=True)
            axs = axs.ravel()
            for month in range(12):
                dat = data[var_i, month_ii[month]]
                dat = dat[np.isfinite(dat)]
                label = "obs"
                if self.sim_sea is not None:
                    label = [label, "sim"]
                    dat = [dat, self.sim_sea_dis[var_i, sim_month_ii[month]]]
                axs[month].hist(
                    dat, bins, density=True, histtype="step", label=label
                )
                if self.sim_sea is None:
                    dat = (dat,)
                ax_cdf = axs[month].twinx()
                for values in dat:
                    values = values[np.isfinite(values)]
                    ranks = (np.arange(len(values)) - 0.5) / len(values)
                    ax_cdf.plot(np.sort(values), ranks)
                    ax_cdf.set_yticklabels([])
                ax_cdf.set_ylim(0, 1)
                axs[month].set_title(month + 1)
            if self.sim_sea is not None:
                axs[0].legend(loc="best")

            suptitle = var_name
            fig.suptitle(suptitle)
            fig.tight_layout(rect=(0, 0, 1, 0.95))
            figs += [fig]
        if len(figs) == 1:
            figs = figs[0]
        for ax in axs:
            ax.grid(True)
        return figs, axs

    def plot_episode_hists(self, var_names=None):
        """Plots histograms of episode duration and amplitude grouped by
        variable."""
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, str):
            var_names = (var_names,)

        figs = []
        axs = []
        for var_name in var_names:
            var_i = self.var_names.index(var_name)
            medians = self.fitted_medians(var_name, self.data_doys)
            devs_obs = (
                self.data_raw[var_i] / self.sum_interval[var_i] - medians
            )
            sign_changes = np.where(np.diff(np.sign(devs_obs)) != 0)[0]
            durations = [np.diff(sign_changes)]
            amplitudes = [
                np.array(
                    [
                        np.mean(epi)
                        for epi in np.split(devs_obs, sign_changes)
                        if len(epi) > 0
                    ]
                )
            ]

            if self.sim_sea is not None:
                medians = self.fitted_medians(var_name, self.sim_doys)
                devs_sim = self.sim_sea[var_i] - medians
                sign_changes = np.where(np.diff(np.sign(devs_sim)) != 0)[0]
                durations_sim = np.diff(sign_changes)
                amplitudes_sim = [
                    np.mean(epi)
                    for epi in np.split(devs_sim, sign_changes)
                    if len(epi) > 0
                ]
                durations += [durations_sim]
                amplitudes += [amplitudes_sim]

            fig, axs_ = plt.subplots(2, 1)
            axs_[0].hist(durations, 30, density=True)
            axs_[0].set_xlabel("Duration [days]")
            axs_[1].hist(amplitudes, 30, density=True)
            axs_[1].set_xlabel("Deviation %s" % conf.units[var_name])

            legend1 = r"observed $\overline{x}=%.2f$" % np.mean(
                durations[0] - 1
            )
            legend2 = r"observed $\sigma=%.2f$" % np.std(
                amplitudes[0][np.isfinite(amplitudes[0])]
            )
            if self.sim_sea is not None:
                legend1 = [
                    legend1,
                    r"sim $\overline{x}=%.2f$" % np.mean(durations_sim - 1),
                ]
                legend2 = [
                    legend2,
                    r"sim $\sigma=%.2f$" % np.std(amplitudes_sim),
                ]
            else:
                legend1, legend2 = (legend1,), (legend2,)
            axs_[0].legend(legend1, loc="best")
            axs_[1].legend(legend2, loc="best")

            suptitle = "Episode statistics %s" % var_name
            fig.suptitle(suptitle)
            try:
                fig.canvas.set_window_title(suptitle)
            except AttributeError:
                pass
            figs += [fig]
            axs += [axs_]
        return figs, axs

    def plot_daily_cycles(self, var_names=None, rain_thresh=None, **fig_kw):
        """Plot cycle of hourly means over the year.

        Climbing Mt. Allon.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        rain_thresh : None, True or float, optional
            Which elemts to include based on rain threshold. If True, it is
            inferred from the configuration file (dist_kwds["R"]["threshold"])
        """
        if self.sim_sea_dis is None:
            print("Call disaggregate first")
            return
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, str):
            var_names = (var_names,)

        if not rain_thresh:
            # rain_mask_obs = np.ones_like(self.met["R"], dtype=bool)
            # rain_mask_sim = np.ones(self.sim_sea_dis.shape[1], dtype=bool)
            rain_comparison = False
        if rain_thresh is True or type(rain_thresh) is float:
            if rain_thresh is True:
                rain_thresh = conf.threshold
            rain_mask_obs = (
                self.data_raw[self.var_names.index("R")] >= rain_thresh
            )
            rain_mask_obs = np.repeat(
                rain_mask_obs, self.sum_interval_dict["R"][0]
            )
            rain_mask_sim = (
                self.sim_sea[self.var_names.index("R")] >= rain_thresh
            )
            rain_mask_sim = np.repeat(
                rain_mask_sim, self.sum_interval_dict["R"][0]
            )
            rain_comparison = True

        figs, axes = [], []
        month_iis = [
            varwg.times.time_part(self.times_orig, "%m") == month
            for month in range(1, 13)
        ]
        hour_iis = [
            varwg.times.time_part(self.times_orig, "%H") == hour
            for hour in range(24)
        ]
        month_sim_iis = [
            varwg.times.time_part(self.dis_times, "%m") == month
            for month in range(1, 13)
        ]
        hour_sim_iis = [
            varwg.times.time_part(self.dis_times, "%H") == hour
            for hour in range(24)
        ]

        fig_kw.update(dict(sharex=True, sharey=True))
        for var_name in var_names:
            if rain_comparison:
                fig, axs = plt.subplots(2, 2, **fig_kw)
            else:
                fig, axs = plt.subplots(1, 2, **fig_kw)
                axs = np.array((axs,))

            var_i_dis = self.var_names.index(var_name)

            for mask_i, mask_mod in enumerate((lambda x: x, lambda x: ~x)):
                if not rain_comparison and mask_i == 1:
                    break

                means_obs = np.zeros((12, 24))
                means_sim = np.zeros_like(means_obs)
                for month in range(1, 13):
                    for hour in range(24):
                        month_ii = month_iis[month - 1]
                        hour_ii = hour_iis[hour]
                        mask = month_ii & hour_ii
                        if rain_comparison:
                            mask &= mask_mod(rain_mask_obs)
                        data_obs = np.array(self.met[var_name])[mask]
                        month_ii = month_sim_iis[month - 1]
                        hour_ii = hour_sim_iis[hour]
                        data_sim = self.sim_sea_dis[var_i_dis]
                        mask = month_ii & hour_ii
                        if rain_comparison:
                            mask &= mask_mod(rain_mask_sim)
                        data_sim = data_sim[mask[: len(data_sim)]]
                        means_obs[month - 1, hour] = np.nanmean(data_obs)
                        means_sim[month - 1, hour] = np.nanmean(data_sim)

                # the following code tries to achieve that the colorbar for the
                # two contour plots spans from vmin to vmax (this would be the
                # case for calls to imshow instead of contourf).
                # see
                # http://matplotlib.1069221.n5.nabble.com/
                # question-about-contours-and-clim-td21111.html
                # and
                # https://github.com/matplotlib/matplotlib/pull/2176
                # to understand why this is so annoying
                vmin = min(means_obs.min(), means_sim.min())
                vmax = max(means_obs.max(), means_sim.max())
                locator = ticker.MaxNLocator(8)
                locator.create_dummy_axis()
                levs = locator.tick_values(vmin, vmax)
                axs[mask_i, 0].contourf(
                    list(range(1, 25)), list(range(1, 13)), means_obs, levs
                )
                co = axs[mask_i, 1].contourf(
                    list(range(1, 25)), list(range(1, 13)), means_sim, levs
                )

                axs[mask_i, 0].set_xticks([6, 12, 18, 24])
                axs[mask_i, 1].set_xticks([6, 12, 18, 24])

                axs[mask_i, 0].set_title("Observed")
                axs[mask_i, 0].set_xlabel("hour")
                axs[mask_i, 0].set_ylabel("month")
                axs[mask_i, 0].grid()
                axs[mask_i, 1].set_title("Simulated")
                axs[mask_i, 1].set_xlabel("hour")
                axs[mask_i, 1].grid()

            title = "Seasonal daily cycles for %s" % conf.ygreek[var_name]
            fig.suptitle(title)
            fig.name = "daily_cycles_%s" % var_name
            try:
                fig.canvas.set_window_title(title)
            except AttributeError:
                pass
            if "constrained_layout" not in fig_kw.keys():
                fig.tight_layout(rect=(0, 0, 1, 0.925))
            fig.subplots_adjust(right=0.8)
            cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
            cbar = fig.colorbar(co, cax=cbar_ax)
            cbar.set_label(conf.units[var_name])

            figs += [fig]
            axes += [axs]
        return figs, axes

    def plot_candidates(self, figs=None, axss=None):
        if not hasattr(self, "candidates"):
            print("No candidates for plotting found.")
            return None, None
        fig, axs = self._meteogram(
            self.sim_times,
            self.sim,
            "Resample candidate bounds",
            var_names=self.var_names,
            plot_dewpoint=False,
            fig=figs,
            axs=axss,
        )
        for candidates, ax in zip(self.candidates, axs):
            mins = np.min(candidates, axis=1)
            q25 = np.quantile(candidates, 0.25, axis=1)
            q75 = np.quantile(candidates, 0.75, axis=1)
            maxs = np.max(candidates, axis=1)
            ax[0].fill_between(
                self.sim_times, mins, maxs, alpha=0.25, color="k"
            )
            ax[0].fill_between(self.sim_times, q25, q75, alpha=0.5, color="k")
        return fig, axs

    # def plot_torus(self, var_names=None, **fig_kw):
    #     """Provides a 2d visualization with doys as horizontal and hours as
    #     vertical axis.

    #     Parameters
    #     ----------
    #     var_names : sequence of str, optional
    #         Which variables to plot. None means plot all.
    #     """
    #     if var_names is None:
    #         var_names = self.var_names_dis
    #     elif type(var_names) == str:
    #         var_names = var_names,

    #     figs, axes = [], []
    #     n_years = self.times_orig[-1].year - self.times_orig[0].year + 1
    #     nrows = int(np.ceil(float(n_years) / 10))
    #     ncols = n_years / nrows
    #     for var_name in var_names:
    #         fig, axs = plt.subplots(nrows, ncols, **fig_kw)
    #         # torus =

    #         figs += [fig]
    #         axes += [axs]


# Backward compatibility alias
VGPlotting = Plotting


if __name__ == "__main__":
    import varwg

    met_vg = varwg.VG(("theta", "Qsw", "ILWR", "rh", "u", "v"), verbose=True)
    met_vg.simulate()
    # met_vg.disaggregate()
    met_vg.plot_autocorr()
    plt.show()



import time
import numpy as np
import pandas as pd
import numpy.testing as npt
import varwg
from varwg.meteo import meteox2y, meteox2y_cy, brunner

import xarray as xr


class Test(npt.TestCase):
    def setUp(self):
        dtimes = pd.date_range("2000", "2001", freq="h")
        self.lat = 47.66
        self.lon = 9.18
        self.doys = varwg.times.datetime2doy(dtimes)

    def test_pot_s_rad(self):
        time0 = time.perf_counter()
        qsw_2d = -1 * np.ones((22 - 4, len(self.doys)))
        for i, doy in enumerate(self.doys):
            date = ["%03d %02d" % (doy, hour) for hour in range(4, 22)]
            qsw_2d[:, i] = meteox2y.pot_s_rad(
                date, longt=self.lon, lat=self.lat, in_format="%j %H"
            )
        qsw_py = qsw_2d.sum(axis=0).ravel()
        time_py = time.perf_counter()

        qsw_2d = -1 * np.ones((22 - 4, len(self.doys)))
        for i, doy in enumerate(self.doys):
            date = ["%03d %02d" % (doy, hour) for hour in range(4, 22)]
            qsw_2d[:, i] = meteox2y_cy.pot_s_rad(
                date, longt=self.lon, lat=self.lat, in_format="%j %H"
            )
        qsw_cy = qsw_2d.sum(axis=0).ravel()
        time_cy = time.perf_counter() - time_py
        time_py -= time0
        print("Time python: ", time_py)
        print("Time cython: ", time_cy)
        npt.assert_almost_equal(qsw_cy, qsw_py)

    def test_sunshine_pot(self):
        time0 = time.perf_counter()
        hours_cy = meteox2y_cy.sunshine_pot(
            self.doys, lat=self.lat, longt=self.lon
        )
        time_cy = time.perf_counter()
        hours_py = meteox2y.sunshine_pot(
            self.doys, lat=self.lat, longt=self.lon
        )
        time_py = time.perf_counter() - time_cy
        time_cy -= time0
        print("Time python: ", time_py)
        print("Time cython: ", time_cy)
        npt.assert_almost_equal(hours_cy, hours_py)

    def test_brunner_compound(self):
        Ta = np.arange(10.0)
        P = Ta[::-1]
        hot_dry_cy = meteox2y_cy.brunner_compound(Ta, P)
        npt.assert_almost_equal(
            hot_dry_cy, [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        )
        hot_dry_py = brunner.brunner_compound(Ta, P)
        npt.assert_almost_equal(hot_dry_cy, hot_dry_py)

        rs = np.random.RandomState(0)
        Ta = xr.DataArray(
            rs.randn(2, 100, 10000), dims=["model", "station", "time"]
        )
        P = Ta + 0.5 * rs.randn(*Ta.shape)
        bc_py = brunner.brunner_compound(Ta, P, sequential=True, progress=True)
        bc_cy = meteox2y_cy.brunner_compound(Ta, P, progress=True)
        npt.assert_almost_equal(bc_py, bc_cy)

    def tearDown(self):
        pass


if __name__ == "__main__":
    npt.run_module_suite()



"""
Created on 06.09.2012

@author: dirk
"""

import matplotlib.pyplot as plt
import numpy as np
from numpy.testing import TestCase, assert_almost_equal

import varwg
from varwg.time_series_analysis import time_series


class Test(TestCase):
    def setUp(self):
        self.verbose = False
        self.data = np.array(
            [
                47,
                64,
                23,
                71,
                38,
                64,
                55,
                41,
                59,
                48,
                71,
                35,
                57,
                40,
                58,
                44,
                80,
                55,
                37,
                74,
                51,
                57,
                50,
                60,
                45,
                57,
                50,
                45,
                25,
                59,
                50,
                71,
                56,
                74,
                50,
                58,
                45,
                54,
                36,
                54,
                48,
                55,
                45,
                57,
                50,
                62,
                44,
                64,
                43,
                52,
                38,
                59,
                55,
                41,
                53,
                49,
                34,
                35,
                54,
                45,
                68,
                38,
                50,
                60,
                39,
                59,
                40,
                57,
                54,
                23,
            ],
            dtype=float,
        )
        self.sim = self.gen_time_series(rho=0.9)

    def gen_time_series(self, rho):
        # time series with known autocorrelation
        self.T = int(1e5)
        sim = varwg.rng.normal(size=self.T)
        for tt in range(self.T):
            sim[tt] += rho * sim[tt - 1]
        return sim

    def test_autocorr_ar(self):
        """compare calculated autocorr on series with known autocorr."""
        phi = time_series.auto_corr(self.sim, 1)
        assert_almost_equal(0.9, phi, decimal=2)

    def test_autocorr_ar_multi(self):
        """compare calculated autocorr on series with known autocorr 2dim."""
        phi = time_series.auto_corr([self.sim, self.gen_time_series(0.6)], 1)
        assert_almost_equal([0.9, 0.6], phi, decimal=2)

    def test_autocorr_ar_nans(self):
        """compare calculated autocorr on series with known autocorr with nans."""
        sim = np.copy(self.sim)
        sim[:: int(self.T / 5)] = np.nan
        sim[1 :: int(self.T / 5)] = np.nan
        phi = time_series.auto_corr(sim, 1)
        assert_almost_equal(0.9, phi, decimal=2)

    def test_autocorr_ar_nans_multi(self):
        """compare calculated autocorr on series with known autocorr with nans
        2dim.
        """
        sim = np.copy([self.sim, self.gen_time_series(0.6)])
        sim[0, :: int(self.T / 3)] = np.nan
        sim[:, 1 :: int(self.T / 5)] = np.nan
        phi = time_series.auto_corr(sim, 1)
        assert_almost_equal([0.9, 0.6], phi, decimal=2)

    def test_autocorr_plt_acorr(self):
        """do we get the same as matplotlib.pyplot.acorr?"""
        corr_exp = plt.acorr(self.sim, maxlags=5)[1]
        corr = time_series.auto_corr(self.sim, range(-5, 6))
        assert_almost_equal(corr_exp, corr, decimal=4)

    def test_autocorr_box1(self):
        """see p.30f Time Series Analysis Box Jenkins"""
        phi = time_series.auto_corr(self.data[:10], 1)
        assert_almost_equal(phi, -0.79, decimal=2)

    def test_autocorr_box2(self):
        """see p.31f Time Series Analysis Box Jenkins"""
        rk = np.array(
            [
                -0.39,
                0.30,
                -0.17,
                0.07,
                -0.10,
                -0.05,
                0.04,
                -0.04,
                -0.00,
                0.01,
                0.11,
                -0.07,
                0.15,
                0.04,
                -0.01,
            ]
        )
        phis = np.array(
            [time_series.auto_corr(self.data, k) for k in range(1, 16)]
        )
        if self.verbose:
            for ii, (r, phi) in enumerate(zip(rk, phis)):
                print("%d %+.4f %+.4f" % (ii + 1, r, phi))
        assert_almost_equal(rk, phis, decimal=2)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_auto_corr']
    # run_module_suite()
    pass



import numpy as np
import numpy.testing as npt
from varwg.time_series_analysis import spectral, time_series, models
from varwg.time_series_analysis.tests import test_models
import varwg

# these are empirical covariances from stdn constance data
# theta, Qsw and ILWR
# cov = np.array([[0.92887129,  0.23603728,  0.58140583],
#                 [0.23603728,  1.01204546, -0.44856305],
#                 [0.58140583, -0.44856305,  0.9826686]])
cov = np.array([[0.93, 0.24, 0.58], [0.24, 1.01, -0.45], [0.58, -0.45, 0.98]])
autocov_data = np.array(
    [
        [0.93, 0.75, 0.54, 0.41, 0.32, 0.25, 0.2, 0.17],
        [1.01, 0.31, 0.12, 0.06, 0.04, 0.02, 0.01, 0.03],
        [0.98, 0.62, 0.34, 0.22, 0.16, 0.11, 0.09, 0.08],
    ]
)
K = cov.shape[0]


def ac(var_i):
    """An autocovariance "model" based on the empirical autocovariances
    above."""

    def ac_(lag):
        lag = np.atleast_1d(lag)
        ii = lag < autocov_data.shape[1]
        covs = np.zeros_like(lag) + 1e-6
        covs[ii] = autocov_data[var_i, lag[ii]]
        return np.squeeze(covs)

    return ac_


autocovs = [ac(0), ac(1), ac(2)]


class Test(npt.TestCase):
    def setUp(self):
        varwg.reseed(3)
        # spectral.simulate is sensitive towards T! low T -> bad fit
        self.T = 25000
        # generating a ar-1 time series which has a well known
        # autocovariance function
        # high rho, worse fit
        self.rho = rho = 0.75
        test_data = varwg.rng.normal(size=self.T)
        for t in range(1, self.T):
            test_data[t] += rho * test_data[t - 1]
        self.test_data = test_data - test_data.mean()
        self.test_data /= self.test_data.std()

    def tearDown(self):
        pass

    def test_1d_simulate(self):
        def cov_model_ar1(lag):
            return self.rho**lag / (1 - self.rho**2)

        n_lags = 10
        cov_exp = cov_model_ar1(np.arange(n_lags))
        spec = spectral.Spectral(cov_model_ar1, self.T, pool_size=2)
        data_sim = spec.sim_n(5)
        for data in data_sim:
            cov_act = np.array(
                [time_series.auto_cov(data, lag) for lag in range(n_lags)]
            )
            npt.assert_almost_equal(cov_act, cov_exp, decimal=1)
        data_sim = spec.sim
        cov_act = np.array(
            [time_series.auto_cov(data_sim, lag) for lag in range(n_lags)]
        )
        npt.assert_almost_equal(cov_act, cov_exp, decimal=1)

    def test_nd_simulate(self):
        def cov_model(h):
            return np.exp(-h)

        domainshape = 500, 500, 10
        # cov_exp = cov_model(np.arange(domainshape[0]))
        spec = spectral.SpectralND(
            cov_model, domainshape, pool_size=2, scale=(10, 5, 1)
        )
        # cov_act =
        data_sim = spec.sim
        npt.assert_array_equal(data_sim.shape, domainshape)
        npt.assert_almost_equal(data_sim.std(), 1.0, decimal=1)
        # can we get a different standard deviation?
        spec.sigma = sigma = 2.5
        npt.assert_almost_equal(spec.sim.std(), spec.sigma, decimal=1)
        spec = spectral.SpectralND(
            cov_model, domainshape, pool_size=2, scale=(10, 5, 1), sigma=sigma
        )
        data_sims = spec.sim_n(5)
        for data_sim in data_sims:
            npt.assert_array_equal(data_sim.shape, domainshape)
            npt.assert_almost_equal(data_sim.std(), spec.sigma, decimal=1)

    def test_2d_simulate(self):
        # def cov_model_ar1(rho):
        #     return lambda lag: rho ** lag / (1 - rho ** 2)
        # these are mild differences in autocorrelation. with more
        # different ones the last test fails
        # autocovs = [cov_model_ar1(.6),
        #             cov_model_ar1(.65),
        #             cov_model_ar1(.75)]
        mspec = spectral.MultiSpectral(autocovs, cov, self.T)
        data_sim = mspec.sim
        cov_sim = np.cov(data_sim)
        npt.assert_almost_equal(cov_sim, cov, decimal=2)
        lags = list(range(1, 7))
        for var_i, var in enumerate(data_sim):
            cov0 = autocovs[var_i](0)
            ac_exp = [autocovs[var_i](lag) / cov0 for lag in lags]
            ac_act = [time_series.auto_corr(var, lag) for lag in lags]
            npt.assert_almost_equal(ac_act, ac_exp, decimal=1)

    def test_spectral_2d_covdata(self):
        """Test if MultiSpectral works with (auto)covs given as array."""
        T = 7500
        data_obs = models.VAR_LS_sim(
            test_models.B_test, test_models.sigma_u_test, T
        )
        data_obs -= data_obs.mean(axis=1)[:, None]
        data_obs /= data_obs.std(axis=1)[:, None]
        mspec = spectral.MultiSpectral(data_obs, data_obs, T)
        data_sim = mspec.sim
        npt.assert_almost_equal(np.cov(data_sim), np.cov(data_obs), decimal=1)
        for var_i in range(data_obs.shape[0]):
            ac_obs = [
                time_series.auto_cov(data_obs[var_i], lag) for lag in range(7)
            ]
            ac_sim = [
                time_series.auto_cov(data_sim[var_i], lag) for lag in range(7)
            ]
            npt.assert_almost_equal(ac_sim, ac_obs, decimal=1)


if __name__ == "__main__":
    npt.run_module_suite()



# Name:        times
# Purpose: several time conversions and stuff, formerly found in
#          time_array_conversion, timestamp or my_globals
#
# Author:      Thomas Pfaff, Magdalena Eder, Dirk Schlabing
#
# Created:     17.11.2011
# Copyright:   (c) guttenberg 2011
# Licence:     <who cares?>
#!/usr/bin/env python
"""
Time helpers and conversions (:mod:`times`)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Helper functions to convert arrays containing time information.
All the functions with a ``2`` in the name provide the advertised conversions
for scalar as well as for *array* input.

.. currentmodule:: varwg.times

.. autosummary::
   :nosignatures:
   :toctree: generated/

    cwr2datetime
    cwr2str
    cwr2unix
    datetime2doy
    datetime2ordinal
    datetime2str
    datetime2unix
    doy2datetime
    iso2datetime
    iso2unix
    ordinal2datetime
    str2datetime
    str2ordinal
    str2unix
    unix2cwr
    unix2datetime
    unix2ordinal
    unix2str

    time_part
    time_part_sort
    timestamp2index
    index2timestamp
"""
from datetime import datetime, timedelta
import itertools
import time
import calendar
import numpy as np
import pandas as pd


class TimeParseError(ValueError):
    pass


def cwr2datetime(cwr_string):
    """Converts a CWR-timesting into a datetime object.
    >>> cwr2datetime("2040001.00")
    datetime.datetime(2040, 1, 1, 0, 0)
    >>> cwr2datetime('1980001')
    datetime.datetime(1980, 1, 1, 0, 0)
    """

    def cwr2datetime_single(cwr_string_single):
        year = int(cwr_string_single[:4])
        doy = int(cwr_string_single[4:7]) - 1
        try:
            decimal_day = float(cwr_string_single[7:])
        except ValueError:
            decimal_day = 0
        seconds = decimal_day * 24 * 60**2
        return datetime(year, 1, 1) + timedelta(days=doy, seconds=seconds)

    try:
        return cwr2datetime_single(str(cwr_string))
    except ValueError:
        cwr_string = np.array(cwr_string).astype(str)
        return np.array([cwr2datetime_single(cwr) for cwr in cwr_string])


def cwr2str(cwr_string, d_format="%d.%m.%Y %H:%M:%S"):
    """Converts a CWR-timestring into a human-readable timestring.

    >>> cwr2str("2040001.00")
    '01.01.2040 00:00:00'
    """
    dt = cwr2datetime(cwr_string)
    return datetime2str(dt, d_format)


def cwr2unix(cwr_string):
    """Converts a CWR-timestring to a unix timestamp.

    >>> cwr2unix("2040001.00")
    2208988800.0
    """

    def cwr2unix_single(cwr_string_single):
        cwr_string_single = "%013.5f" % float(cwr_string_single)
        yearday = int(float(cwr_string_single[4:]))  # round down to whole days
        # this saves us from calculating the month and day-of-month
        timestamp = str2unix(cwr_string_single[:7], "%Y%j")
        seconds = (float(cwr_string_single[4:]) - yearday) * 86400
        return timestamp + seconds

    try:
        return cwr2unix_single(cwr_string)
    except (ValueError, TypeError):
        return np.array([cwr2unix_single(cwr) for cwr in cwr_string])


def dy2datetime(dy):
    """Converts a DYRESM float time (a.k.a. julian date) into a datetime
    object.

    >>> dy2datetime(2451544.5)
    datetime.datetime(2000, 1, 1, 0, 0)
    >>> dy2datetime(2456462.4375)
    datetime.datetime(2013, 6, 18, 22, 30)
    """
    return ordinal2datetime(dy - 1721424.5)


def datetime2ordinal(dt):
    """Converts a datetime object into an ordinal.

    >>> import datetime
    >>> dt = datetime.datetime(2040, 1, 1, 0, 0, 0, 500000)
    >>> datetime2ordinal(dt)
    744730
    >>> datetime2ordinal(datetime.datetime(1, 1, 1, 0, 0))
    1
    """
    try:
        return np.array([sub_dt.toordinal() for sub_dt in dt])
    except TypeError:
        return dt.toordinal()


def datetime2cwr(dt):
    """Converts a datetime object into a CWR/Julian timestamp.

    >>> import datetime
    >>> datetime2cwr(datetime.datetime(2040, 1, 1, 12, 30, 30, 0))
    2040001.5211805555
    """

    def datetime2cwr_single(dt):
        timetuple = dt.timetuple()
        year = timetuple.tm_year
        yearday = timetuple.tm_yday
        seconds_as_days = (
            float(
                timetuple.tm_hour * 3600
                + timetuple.tm_min * 60
                + timetuple.tm_sec
            )
            / 86400
        )
        return float(
            str(year) + ("%03d" % yearday) + "." + str(seconds_as_days)[2:]
        )

    try:
        return np.array([datetime2cwr_single(sub_dt) for sub_dt in dt])
    except TypeError:
        return datetime2cwr_single(dt)


def date2jdn(dates):
    """Converts date object to julian day number (without hours!)"""

    def date2jdn_single(date):
        a = (14 - date.month) // 12
        y = date.year + 4800 - a
        m = date.month + 12 * a - 3

        return (
            date.day
            + (153 * m + 2) // 5
            + 365 * y
            + y // 4
            - y // 100
            + y // 400
            - 32045
        )

    try:
        return np.array([date2jdn_single(date) for date in dates])
    except TypeError:
        return date2jdn_single(dates)


def datetime2cwr_old(dt):
    return unix2cwr(datetime2unix(dt))


def datetime2str(dt, d_format="%d.%m.%Y %H:%M:%S"):
    """Converts a datetime object into a human-readable string.

    >>> import datetime
    >>> datetime2str(datetime.datetime(2040, 1, 1, 0, 0, 0, 500000))
    '01.01.2040 00:00:00'
    """
    try:
        return np.array(
            [sub_datetime.strftime(d_format) for sub_datetime in dt]
        )
    except TypeError:
        return dt.strftime(d_format)


def datetime2unix(dt):
    """Converts a dateime object into a unix-timestamp.

    >>> import datetime
    >>> dt = datetime.datetime(2040, 1, 1, 0, 0, 0, 500000)
    >>> datetime2unix(dt)
    2208988800.5
    >>> datetime2unix(datetime.datetime(1970, 1, 1, 0, 0))
    0.0
    """
    try:
        tzinfo = dt.tzinfo if hasattr(dt, "tzinfo") else dt[0].tzinfo
    except IndexError:
        tzinfo = None
    diff = dt - datetime(1970, 1, 1, 0, 0, tzinfo=tzinfo)
    try:
        return np.array(
            [
                sub_diff.days * 86400
                + sub_diff.seconds
                + sub_diff.microseconds / 1e6
                for sub_diff in diff
            ]
        )
    except TypeError:
        return diff.days * 86400 + diff.seconds + diff.microseconds / 1e6


def datetime2doy(dt):
    """Extracts the day of year as a float from the given datetimes.

    >>> import datetime
    >>> dt = datetime.datetime(2040, 1, 1, 12, 30, 30, 500000)
    >>> datetime2doy(dt)
    1.5211863425925927
    """

    def datetime2doy_single(dt):
        return (
            dt.timetuple().tm_yday
            + dt.hour / 24.0
            + dt.minute / (24.0 * 60)
            + dt.second / (24.0 * 60**2)
            + dt.microsecond / (24.0 * 60**2 * 1e6)
        )

    try:
        return np.array([datetime2doy_single(sub_dt) for sub_dt in dt])
    except TypeError:
        return datetime2doy_single(dt)


def datetime2hour(dt):
    """Extracts the hour of a day as a float from the given datetimes.

    >>> import datetime
    >>> dt = datetime.datetime(2040, 1, 1, 12, 30)
    >>> datetime2hour(dt)
    12.5
    """

    def datetime2hour_single(dt):
        return (
            dt.hour
            + dt.minute / 60.0
            + dt.second / (60.0**2)
            + dt.microsecond / (60.0**2 * 1e6)
        )

    try:
        return np.array([datetime2hour_single(sub_dt) for sub_dt in dt])
    except TypeError:
        return datetime2hour_single(dt)


def doy2datetime(doy, year=2000):
    """Constructs a datetime object from given day of year.

    Parameters
    ----------
    year : int, optional
        The year of the day of the year.

    Examples
    --------
    >>> import datetime, numpy
    >>> doy2datetime(1.5211863425925927, 2040)
    datetime.datetime(2040, 1, 1, 12, 30, 30, 500000)
    >>> doy2datetime(numpy.arange(1,3))
    array([datetime.datetime(2000, 1, 1, 0, 0),
           datetime.datetime(2000, 1, 2, 0, 0)], dtype=object)
    """

    def doy2datetime_single(doy, year):
        return datetime(int(year), month=1, day=1) + timedelta(
            days=float(doy) - 1
        )

    # TODO: interpret 365, 0 as a new year
    try:
        if type(year) is int:
            years = itertools.cycle([year])
        else:
            years = year
        return np.array(
            [
                doy2datetime_single(sub_doy, sub_year)
                for sub_doy, sub_year in zip(doy, years)
            ]
        )
    except TypeError:
        return doy2datetime_single(doy, year)


# these two are virtually the same
def iso2datetime(iso):
    """Converts an ISO formated time string to a datetime object."""
    try:
        return str2datetime(iso, "%Y-%m-%dT%H:%M:%S.%f")
    except (ValueError, TypeError):
        return str2datetime(iso, "%Y-%m-%dT%H:%M:%S")


def datetimefromisoformat(ts, fmt="%Y-%m-%dT%H:%M:%S"):
    return datetime.strptime(ts, fmt)


# (same same but different)


def iso2unix(iso):
    """Converts an ISO formated time string to a unix time stamp."""
    try:
        return str2unix(iso, "%Y-%m-%dT%H:%M:%S.%f")
    except (ValueError, TypeError):
        # sometimes the microsecond is missing
        return str2unix(iso, "%Y-%m-%dT%H:%M:%S")


def ordinal2datetime(ord_):
    """Converts an ordinal to a datetime object.

    >>> ordinal2datetime(744730)
    datetime.datetime(2040, 1, 1, 0, 0)
    """
    try:
        return np.array(
            [
                (
                    datetime.fromordinal(int(sub_ord))
                    + timedelta(sub_ord - int(sub_ord))
                )
                for sub_ord in ord_
            ]
        )
    except TypeError:
        return datetime.fromordinal(int(ord_)) + timedelta(ord_ - int(ord_))


def str2datetime(str_, d_format="%d.%m.%Y %H:%M:%S"):
    """Converts a human readable time string tinto a datetime object.

    >>> str2datetime("01.01.2040 00:00:00")
    datetime.datetime(2040, 1, 1, 0, 0)
    """
    try:
        if not isinstance(str_[0], str):
            raise TypeError("Sequence is of type %s, not str" % type(str_[0]))
        pd_series = pd.to_datetime(
            str_,
            format=d_format,
            # infer_datetime_format=True
        )
        return pd_series.to_pydatetime()
    except (ValueError, AttributeError):
        return datetime.strptime(str_, d_format)


def str2ordinal(str_, d_format="%d.%m.%Y %H:%M:%S"):
    """Converts a human readable time string to an ordinal.

    >>> str2ordinal("01.01.2040 00:00:00")
    744730
    >>> str2ordinal("01.01.0001 00:00:00")
    1
    """
    return datetime2ordinal(str2datetime(str_, d_format))


def str2unix(str_, d_format="%d.%m.%Y %H:%M:%S"):
    """Converts a time-string of the given d_format (default:
    "dd.mm.yyyy HH:MM:SS") into a unix-timestamp.

    >>> str2unix("01.01.2040 00:00:00")
    2208988800.0
    >>> str2unix("01.01.1970 00:00:00")
    0.0
    """
    return datetime2unix(str2datetime(str_, d_format))


def unix2cwr(timestamp):
    """Converts a unix-timestamp to a CWR-timestring.

    >>> unix2cwr(2208988800.0)
    2040001.0
    """

    def unix2cwr_single(timestamp_single):
        src_tuple = time.gmtime(timestamp_single)
        # we are looking for "{year}{yearday}.{seconds expressed in days}
        year = str(src_tuple[0])
        yearday = time.strftime("%j", src_tuple)
        year_yearday_tuple = time.strptime(year + yearday, "%Y%j")
        diff_seconds = timestamp_single - calendar.timegm(year_yearday_tuple)
        seconds_as_days = float(diff_seconds) / 86400
        return "%s%03d%s" % (
            year,
            int(yearday),
            str("%f" % seconds_as_days)[1:],
        )

    try:
        return float(unix2cwr_single(timestamp))
    except TypeError:
        return np.array([float(unix2cwr_single(ts)) for ts in timestamp])


def unix2datetime(timestamp):
    """Convert a unix time stamp to a datetime object.

    >>> import datetime
    >>> unix2datetime(2208988800.5)
    datetime.datetime(2040, 1, 1, 0, 0, 0, 500000)
    >>> unix2datetime(0)
    datetime.datetime(1970, 1, 1, 0, 0)
    """
    try:
        return np.array(
            [
                datetime(1970, 1, 1) + timedelta(seconds=float(stamp))
                for stamp in timestamp
            ]
        )
    except TypeError:
        return datetime(1970, 1, 1) + timedelta(seconds=float(timestamp))


def unix2ordinal(timestamp):
    """Converts a unix time stamp to an ordinal.

    >>> unix2ordinal(2208988800.0)
    744730
    """
    try:
        np.array(
            [datetime2ordinal(unix2datetime(stamp)) for stamp in timestamp]
        )
    except TypeError:
        return datetime2ordinal(unix2datetime(timestamp))


def unix2str(timestamp, d_format="%d.%m.%Y %H:%M:%S"):
    """Converts a unix-timestamp into a time-string of the given d_format
    (default: "dd.mm.yyyy HH:MM:SS").

    >>> unix2str(2208988800.5)
    '01.01.2040 00:00:00'
    >>> unix2str(0)
    '01.01.1970 00:00:00'
    """
    return datetime2str(unix2datetime(timestamp), d_format)


def xls2datetime(xldate, datemode=0):
    """Here's the bare-knuckle no-seat-belts use-at-own-risk version
    posted by John Machin in http://stackoverflow.com
    datemode: 0 for 1900-based, 1 for 1904-based
    """

    def xls2datetime_single(xldate, datemode):
        return (
            # 30.Dez, weil 1900 in Excel ein Schaltjahr ist --> ab dem 1. Maerz
            # stimmts
            datetime(1899, 12, 30)
            + timedelta(days=xldate + 1462 * datemode)
        )

    try:
        return xls2datetime_single(xldate, datemode)
    except (ValueError, TypeError):
        return np.array([xls2datetime_single(xl, datemode) for xl in xldate])


def timedelta2seconds(dt):
    return dt.days * 86400 + dt.seconds + dt.microseconds * 1e-6


def timedelta2slice(delta, dt, offset=0, step=None):
    start = offset
    stop = (
        None
        if delta is None
        else offset + int(timedelta2seconds(delta) / timedelta2seconds(dt))
    )
    return slice(start, stop, step)


def timestamps2slice(startts=None, endts=None, dt=None, refts=None, step=None):
    """Ask Thomas.

    >>> timestamps2slice('2010-03-01T23:54:00', '2010-03-07T23:52:00', \
                         timedelta(days=1))
    slice(0, 5, None)
    >>> timestamps2slice('2010-03-01T23:54:00', '2010-03-07T23:52:00', \
                         timedelta(days=1), '2010-03-04T00:00:00')
    slice(-2, 3, None)
    >>> timestamps2slice('2010-03-01T23:54:00', '2010-03-07T23:52:00', \
                         timedelta(days=1), '2010-01-04T00:00:00')
    slice(56, 61, None)
    """
    # Dirk: Ist das Kunst oder kann das weg?
    #    if dt is None:
    #        tdelta = timedelta(seconds=1)

    if startts is None:
        tstart = None
    if type(startts) == str:
        tstart = datetimefromisoformat(startts)
    else:
        tstart = startts

    if endts is None:
        tend = None
    if type(endts) == str:
        tend = datetimefromisoformat(endts)
    else:
        tend = endts

    if refts is None:
        tref = tstart
    elif type(refts) == str:
        tref = datetimefromisoformat(refts)
    else:
        tref = refts

    if (tstart is None) or (tref is None):
        slicestart = None
    else:
        slicestart = int(
            timedelta2seconds(tstart - tref) / timedelta2seconds(dt)
        )

    return timedelta2slice(
        None if ((tend is None) or (tstart is None)) else tend - tstart,
        dt,
        slicestart,
        step,
    )


def timestamp2index(ts, dt, refts, **kwargs):
    """Calculates the array index for a certain time in an equidistant
    time-series given the reference time (where the index would be 0)
    and the time discretization.
    If any of the input parameters contains timezone information, all others
    also need to contain timezone information.

    Parameters
    ----------
    ts        : str or datetime-object
                The timestamp to determine the index for
                If it is a string, it will be converted to datetime using the
                function _datetimefromisoformat Formatting keywords may be
                passed to this function

    dt        : str or timedelta object
                The discretization of the time series (the amount of time that
                elapsed between indices)
                If used as a string, it needs to be given in the format
                "keyword1=value1,keyword2=value2". Keywords must be understood
                by the timedelta constructor (like days, hours,
                minutes, seconds) and the values may only be integers.

    refts     : str or datetime-object
                The timestamp to determine the index for
                If it is a string, it will be converted to datetime using the
                function _datetimefromisoformat Formatting keywords may be
                passed to this function

    Returns
    -------
    index    : integer
               The index of a discrete time series array of the given
               parameters

    Examples
    --------
    >>> timestr1, timestr2 = '2008-06-01T00:00:00', '2007-01-01T00:00:00'
    >>> timestamp2index(timestr1, 'minutes=5', timestr2)
    148896
    >>> timestamp2index(timestr1, 'hours=1,minutes=5',timestr2)
    11453
    >>> timestamp2index(timestr1, timedelta(hours=1, minutes=5), timestr2)
    11453
    """
    if not isinstance(ts, datetime):
        _ts = datetimefromisoformat(ts, **kwargs)
    else:
        _ts = ts
    if not isinstance(refts, datetime):
        _refts = datetimefromisoformat(refts, **kwargs)
    else:
        _refts = refts
    if not isinstance(dt, timedelta):
        kwargs = dict(
            [
                (sp[0], int(sp[1]))
                for sp in [item.split("=") for item in dt.split(",")]
            ]
        )
        _dt = timedelta(**kwargs)
    else:
        _dt = dt
    return int(timedelta2seconds(_ts - _refts) / timedelta2seconds(_dt))


def isoformat2unix(ts):
    dt = timedelta(seconds=1)
    tstart = datetimefromisoformat("1970-01-01T00:00:00")
    return timestamp2index(datetimefromisoformat(ts), dt, tstart)


def index2timestamp(idx, dt, refts, **kwargs):
    """Calculates the ISOstring timestamp for a certain index in an equidistant
    time-series given the reference time (where the index would be 0)
    and the time discretization.
    If any of the input parameters contains timezone information, all others
    also need to contain timezone information.

    Parameters
    ----------
    idx        : str or datetime-object
                The timestamp to determine the index for
                If it is a string, it will be converted to datetime using the
                function _datetimefromisoformat Formatting keywords may be
                passed to this function

    dt        : str or timedelta object
                The discretization of the time series (the amount of time that
                elapsed between indices)
                If used as a string, it needs to be given in the format
                "keyword1=value1,keyword2=value2". Keywords must be understood
                by the timedelta constructor (like days, hours,
                minutes, seconds) and the values may only be integers.

    refts     : str or datetime-object
                The timestamp to determine the index for
                If it is a string, it will be converted to datetime using the
                function _datetimefromisoformat Formatting keywords may be
                passed to this function

    Returns
    -------
    ISO-timestamp    : string
               The ISO-timestamp of a discrete time series array of the given
               parameters in this format: '%Y-%m-%dT%H:%M:%S'

    Examples
    --------
    >>> index2timestamp(25637, 'seconds=10', '2010-09-25T00:00:10')
    '2010-09-27T23:13:00'
    >>> index2timestamp(365, 'days=1', '2010-09-25T00:00:10')
    '2011-09-25T00:00:10'
    """
    if not isinstance(refts, datetime):
        _refts = datetimefromisoformat(refts, **kwargs)
    else:
        _refts = refts
    if not isinstance(dt, timedelta):
        kwargs = dict(
            [
                (sp[0], int(sp[1]))
                for sp in [item.split("=") for item in dt.split(",")]
            ]
        )
        _dt = timedelta(**kwargs)
    else:
        _dt = dt

    return str(datetime.isoformat(_refts + idx * _dt))


# Time functions from Dirk's my_globals:
def build_diff_timestring(diff):
    """Split time difference given in seconds into a string representation of
    days, hours, minutes and seconds.
    """
    days = hours = minutes = 0
    seconds = abs(diff)
    while seconds >= 60:
        minutes += 1
        seconds -= 60
    while minutes >= 60:
        hours += 1
        minutes -= 60
    while hours >= 24:
        days += 1
        hours -= 24

    diff_str = ""
    if days > 0:
        diff_str += "%d d " % days
    if hours > 0:
        diff_str += "%d h " % hours
    if minutes > 0:
        diff_str += "%d m " % minutes
    diff_str += "%f s" % seconds

    return diff_str


def time_part(timestamps_or_datetimes, sub_format_str):
    """Returns the "time_part" of a timestamp or datetimes.
    The time_part has to be convertible into an integer.
    This is useful to group values according to months, weeks and so on.
    """

    def single_time_part_unix(stamp, d_format):
        return int(unix2str(stamp, d_format)) if np.isfinite(stamp) else None

    def single_time_part_date(time_, d_format):
        return int(time_.strftime(d_format))

    def single_time_part_np(time_, d_format):
        # this feels over-complicated
        return int(
            iso2datetime(np.datetime_as_string(time_)).strftime(d_format)
        )

    times_ = np.asarray(timestamps_or_datetimes)

    if times_.dtype == object:
        single_time_part = single_time_part_date
    elif np.issubdtype(times_.dtype, np.datetime64):
        single_time_part = single_time_part_np
    else:
        single_time_part = single_time_part_unix

    try:
        return np.array(
            [
                single_time_part(timestamp, sub_format_str)
                for timestamp in times_
            ]
        )
    except TypeError:
        return single_time_part(np.asscalar(times_), sub_format_str)


def time_part_(datetimes, date_part):
    try:
        return np.array([getattr(date, date_part) for date in datetimes])
    except TypeError:
        return getattr(datetimes, date_part)


def time_part_sort_(datetimes, values, date_part):
    assert len(datetimes) == len(values)
    time_of_values = np.array([getattr(date, date_part) for date in datetimes])
    grouped_values = []
    times = sorted(set(time_of_values))
    for time_key in times:
        grouped_values += [values[np.where(time_of_values == time_key)[0]]]
    return times, grouped_values


def time_part_sort(timestamps, values, sub_format_str):
    """Groups "values" as a nested list according to the "sub_format_str" of
    the "timestamps".

    >>> dtimes = datetime(2000, 1, 1) + np.arange(4) * timedelta(days=16)
    >>> values = np.arange(4)
    >>> time_part_sort(dtimes, values, "%m")
    ([1, 2], [array([0, 1]), array([2, 3])])
    """
    assert len(timestamps) == len(values)
    time_of_values = time_part(timestamps, sub_format_str)
    grouped_values = []
    times = sorted(set(time_of_values))
    for time_key in times:
        grouped_values += [values[np.where(time_of_values == time_key)[0]]]
    return times, grouped_values


def regularize(values, dtimes, nan=False, main_diff=None):
    """Regularize an irregular time series by linear interpolation.
    The time interval is guessed from the most frequent interval in `times`.

    >>> dtimes = datetime(2000, 1, 1) + np.arange(4) * timedelta(hours=1)
    >>> dtimes[1] += timedelta(minutes=30)
    >>> values = np.arange(4.)
    >>> values[1] = 1.5
    >>> values
    array([ 0. ,  1.5,  2. ,  3. ])
    >>> dtimes
    array([datetime.datetime(2000, 1, 1, 0, 0),
           datetime.datetime(2000, 1, 1, 1, 30),
           datetime.datetime(2000, 1, 1, 2, 0),
           datetime.datetime(2000, 1, 1, 3, 0)], dtype=object)
    >>> regularize(values, dtimes)
    (array([ 0.,  1.,  2.,  3.]), array([datetime.datetime(2000, 1, 1, 0, 0),
           datetime.datetime(2000, 1, 1, 1, 0),
           datetime.datetime(2000, 1, 1, 2, 0),
           datetime.datetime(2000, 1, 1, 3, 0)], dtype=object))
    """
    in_unix = datetime2unix(dtimes)
    t_diffs = np.diff(in_unix)
    unique_t_diffs = np.unique(t_diffs)
    if len(unique_t_diffs) == 1:
        # our work is not needed
        return values, dtimes
    if main_diff is None:
        hist, edges = np.histogram(t_diffs, unique_t_diffs)
        main_diff_seconds = edges[np.argmax(hist)]
        main_diff = timedelta(seconds=main_diff_seconds)
    elif type(main_diff) is timedelta:
        main_diff_seconds = main_diff.total_seconds()
    else:
        main_diff_seconds = main_diff
        main_diff = timedelta(seconds=main_diff_seconds)
    n_diff = int(np.ceil((in_unix[-1] - in_unix[0]) / main_diff_seconds)) + 1
    out_dtimes = dtimes[0] + main_diff * np.arange(n_diff)
    out_unix = datetime2unix(out_dtimes)
    out_values = np.interp(out_unix, in_unix, values)
    if nan:
        # put in nans where there are missing timesteps
        missing = np.where(t_diffs > main_diff_seconds)[0]
        if len(missing) > 1:
            mask = np.zeros_like(out_values, dtype=bool)
            for ii in missing:
                start_dt, end_dt = dtimes[ii : ii + 2]
                mask |= (out_dtimes > start_dt) & (out_dtimes < end_dt)
            out_values[mask] = np.nan
    return out_values, out_dtimes


def expand_timeseries(timestamps, repeats=4, values=None):
    """Interpolates timestamps and repeats according values.  Assumes constant
    length of time-step.

    Examples
    --------
    >>> import numpy as np
    >>> timestamps = str2unix(["01.01.1980 00:00:00", "02.01.1980 00:00:00"])
    >>> values = np.array([1, 2])
    >>> timestamps, values = expand_timeseries(timestamps, 2, values)
    >>> print (unix2str(timestamps), values)
    ['01.01.1980 00:00:00' '01.01.1980 12:00:00' '02.01.1980 00:00:00'
     '02.01.1980 12:00:00'] [1 1 2 2]
    """
    dt = (timestamps[1] - timestamps[0]) / repeats
    dts = np.arange(0, repeats * dt, dt)
    old_len = len(timestamps)
    timestamps = timestamps.repeat(repeats).reshape((old_len, repeats))
    timestamps += dts
    timestamps = timestamps.ravel()
    if values is not None:
        return timestamps, values.repeat(repeats)
    return timestamps


def periodic_distance(x1, x2, period):
    """

    Examples
    --------
    >>> periodic_distance(0, 23, 24)
    array(1.0)
    """
    period = float(period)
    dist = (x1 - x2) % period
    return np.where(dist > period / 2, period - dist, dist)


def doy_distance(doy1, doy2):
    """

    Examples
    --------
    >>> import numpy as np
    >>> doy_distance(0, np.array([364, 0, 1]))
    array([ 1.,  0.,  1.])
    """
    return periodic_distance(doy1, doy2, 365)


def hour_distance(hour1, hour2):
    return periodic_distance(hour1, hour2, 24)


def daily_ranges(dtimes, data):
    """Daily max-min.

    Parameters
    ----------
    dtimes : sequence of datetime objects
        For the time being it is assumed that time steps are equally spaced!
    data : 1d array

    Returns
    -------
    ranges : 1d array
        length is number of days in dtimes.
    """
    step_length = (dtimes[1] - dtimes[0]).total_seconds()
    steps_per_day = 60**2 * 24 / step_length
    data_2d = data.reshape(-1, steps_per_day)
    maxs = np.nanmax(data_2d, axis=1)
    mins = np.nanmin(data_2d, axis=1)
    return maxs - mins


def feb29_mask(dtimes):
    """Returns a mask indicating the location of the additional day in the
    leap years.

    Should help you to get rid off those buggers.

    Parameters
    ----------
    dtimes : ndarray

    Returns
    -------
    mask : boolean ndarray
    """
    months = time_part(dtimes, "%m")
    days = time_part(dtimes, "%d")
    return (months == 2) & (days == 29)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python
"""
Provides functions to plot windroses when given wind directions.
"""
import datetime
import itertools
import numpy as np
import matplotlib.pyplot as plt


# -----------------------------------------------------------------------------#
def _time_part_sort(datetimes, values, sub_format_str):
    """Groups "values" as a nested list according to the "sub_format_str" of
    the "datetimes"."""
    assert len(datetimes) == len(
        values
    ), "datetimes and values have unequal length (%d and %d)" % (
        len(datetimes),
        len(values),
    )
    time_of_values = np.array(
        [dt.strftime(sub_format_str) for dt in datetimes]
    )
    grouped_values = []
    times = sorted(set(time_of_values))
    for time_key in times:
        grouped_values += [values[np.where(time_of_values == time_key)]]
    return times, grouped_values


# -----------------------------------------------------------------------------#
def _yscale_subplots(fig=None, per_type=False, regrid=False):
    """Sets a common y-scale to all subplots.  If per_type is set to True,
    y-scales are distinguished by the type of the subplots."""
    if fig is None:
        fig = plt.gcf()

    if per_type:
        key_func = type
    else:
        key_func = lambda x: "the one to rule them all"

    # see http://matplotlib.sourceforge.net/faq/howto_faq.html#\
    # find-all-objects-in-figure-of-a-certain-type
    ylim_getable = lambda sub: hasattr(sub, "get_ylim")
    ylim_setable = lambda sub: hasattr(sub, "set_ylim")

    # find the y-limits of each subplot
    ymins, ymaxs = {}, {}
    for subplot in fig.findobj(ylim_getable):
        ymin, ymax = subplot.get_ylim()
        sub_type = key_func(subplot)
        if sub_type not in ymins:
            ymins[sub_type], ymaxs[sub_type] = [], []
        ymins[sub_type].append(ymin)
        ymaxs[sub_type].append(ymax)

    # find the extremes for each type of subplot
    ymin, ymax = {}, {}
    for sub_type in list(ymins.keys()):
        ymin[sub_type] = min(ymins[sub_type])
        ymax[sub_type] = max(ymaxs[sub_type])

    # set ylims (and reset grids)
    for subplot in fig.findobj(ylim_setable):
        sub_type = key_func(subplot)
        subplot.set_ylim(ymin=ymin[sub_type], ymax=ymax[sub_type])
        if regrid:
            subplot.set_rgrids(np.linspace(1e-6, ymax[sub_type], 10))


# -----------------------------------------------------------------------------#
def _wind_direction_theta(classes):
    """Generates classes for directions suitable for polar axes plotting.
    Abstraction is a pain here and documenting is also hard, so sorry."""
    classes = np.where(classes > 360, classes - 360, classes)

    # avoid that a bar wraps the wrong way around the arc by cutting out the
    # overlapping class, in other words:
    # i'm a genie in a bottle, you gotta wrap me the right way
    overlap_i = int(np.argwhere(classes[:-1] - classes[1:] > 0))
    classes = np.r_[classes[:overlap_i], classes[overlap_i + 1 :]]
    theta = np.pi * classes / 180
    # fill up the gap with two classes
    class_width = abs(classes[0] - classes[1])
    classes_gap = np.array(
        [359.999 - class_width / 4, 360.001 + class_width / 4]
    )
    theta_gap = np.pi * classes_gap / 180
    return classes, theta, theta_gap, overlap_i


# -----------------------------------------------------------------------------#
def mind_the_gap(hist, overlap_i):
    hist_gap = np.r_[hist[overlap_i], hist[overlap_i]]
    hist = np.r_[hist[:overlap_i], hist[overlap_i + 1 :]]
    return hist, hist_gap


# -----------------------------------------------------------------------------#
def _plot_speed_hist(
    speed_hist,
    speed_bins,
    bar_func,
    theta,
    theta_gap,
    theta_width,
    overlap_i,
    *args,
    **kwds
):
    # "transpose" the speed_hists as they can be easily plotted speed-class
    # and not sector-wise
    speed_hist = speed_hist.T
    speed_hist[np.isnan(speed_hist)] = 0
    speed_cdf = np.cumsum(speed_hist, axis=0)
    lowers = np.zeros_like(speed_cdf)
    lowers[1:] = speed_cdf[:-1]
    for ii, (lower, upper) in enumerate(zip(lowers, speed_hist)):
        hist, hist_gap = mind_the_gap(upper, overlap_i)
        l_hist, l_hist_gap = mind_the_gap(lower, overlap_i)
        label = "<= %.2f m/s" % speed_bins[ii + 1]
        bars = bar_func(
            theta[1:],
            hist,
            theta_width * 0.75,
            bottom=l_hist,
            align="center",
            label=label,
            *args,
            **kwds
        )
        bars_gap = bar_func(
            theta_gap,
            hist_gap,
            0.5 * theta_width,
            bottom=l_hist_gap,
            align="center",
            *args,
            **kwds
        )
        for bar_ in itertools.chain(bars, bars_gap):
            bar_.set_facecolor(plt.cm.jet(ii * len(speed_bins) * 10))
            bar_.set_alpha(0.5 if "alpha" not in kwds else kwds["alpha"])
    plt.legend(
        loc="center right", bbox_to_anchor=(1.55, 0.5), borderaxespad=0.0
    )


# -----------------------------------------------------------------------------#
def _plot_dir_hist(
    hist,
    speed_hist=None,
    speed_bins=None,
    fig=None,
    subfig=False,
    ticks=True,
    *args,
    **kwds
):
    """Helps to combine common ground between windrose and
    _dir_hist_diff_generic.
    Can also be used to plot a windrose if the histogram is already known. But
    in this case the hist has to be passed as hist[::-1]!
    """
    # we know the wind directions have to be in the range (90,450], so we
    # do not use the classes that were returned by np.histogram
    classes = np.linspace(90, 450, len(hist) + 1)
    classes, theta, theta_gap, overlap_i = _wind_direction_theta(classes)

    try:
        overlap_i = overlap_i[0]
    except (KeyError, TypeError):
        pass
    hist, hist_gap = mind_the_gap(hist, overlap_i)

    # polar axes of subfigs are defined when the subfig is created and not
    # here.
    if not subfig:
        if fig is None:
            fig = plt.figure()
        ax = fig.add_axes(
            [0.1 if speed_hist is None else -0.04, 0.1, 0.8, 0.8], polar=True
        )
        bar_func = ax.bar
    else:
        bar_func = plt.bar
    theta_width = abs(theta[1] - theta[0])
    if speed_hist is None:
        bars = bar_func(
            theta[:-1],
            hist,
            theta_width * 0.75,
            bottom=0.0,
            align="center",
            *args,
            **kwds
        )
        bars_gap = bar_func(
            theta_gap,
            hist_gap,
            0.5 * theta_width,
            bottom=0.0,
            align="center",
            *args,
            **kwds
        )
    else:
        _plot_speed_hist(
            speed_hist,
            speed_bins,
            bar_func,
            theta,
            theta_gap,
            theta_width,
            overlap_i,
            *args,
            **kwds
        )

    if speed_hist is None:
        # set colors according to frequency/speed
        all_hists = list(itertools.chain(hist, hist_gap))
        all_bars = itertools.chain(bars, bars_gap)
        rel_hists = [hist / float(sum(all_hists)) for hist in all_hists]
        for freq, bar_ in zip(rel_hists, all_bars):
            bar_.set_facecolor(plt.cm.jet(freq * 10))
            bar_.set_alpha(0.5 if "alpha" not in kwds else kwds["alpha"])

    # we transformed the classes, so we also have to lie about the labels
    xicks = plt.xticks()[0]
    if ticks:
        xlabels = list(range(90, -1, -45)) + list(range(315, 90, -45))
        xlabels = ("E", "NE", "N", "NW", "W", "SW", "S", "SE")
    #        yicks = plt.yticks()[0]
    #        plt.yticks(yicks, [str(ytick) for ytick in yicks])
    else:
        xlabels = [""] * len(xicks)
    yicks = plt.yticks()[0]
    plt.yticks(yicks, [""] * len(yicks))
    plt.xticks(xicks, xlabels)
    return fig


# -----------------------------------------------------------------------------#
def seasonal_windroses(
    datetimes,
    wind_dirs,
    n_sectors=32,
    time_format_str="%m",
    fig=None,
    speed=None,
    title=None,
    normalize=True,
    separate=False,
    *args,
    **kwds
):
    """Plots a figure with subplots of windroses.
    Each subplot is a windrose containing only directions from the measurement
    times signified by "time_format_str". The default ("%m") plots one windrose
    for each distinct month."""
    fig = plt.figure() if fig is None else fig
    times, sorted_values = _time_part_sort(
        datetimes, wind_dirs, time_format_str
    )
    if speed is None:
        sorted_speeds = [None] * len(times)
    else:
        times, sorted_speeds = _time_part_sort(
            datetimes, speed, time_format_str
        )

    # find out how to best fit the subplots in a wide rectangle
    fig_width = int(np.sqrt(len(times))) + 1
    fig_height = fig_width
    while fig_width * (fig_height - 1) >= len(times):
        fig_height -= 1

    # dummy to get a colorbar
    full_subplot = plt.subplot(111, polar=True)
    windrose(
        wind_dirs,
        n_sectors,
        fig=full_subplot,
        subfig=True,
        speed=speed,
        *args,
        **kwds
    )
    if title is not None:
        plt.suptitle(title)

    # needed to fit to a common scale
    subplots = []
    for sub_i, (time, wind_dir, speed) in enumerate(
        zip(times, sorted_values, sorted_speeds)
    ):
        if separate:
            windrose(wind_dir, n_sectors, speed=speed)
        else:
            subplots.append(
                plt.subplot(fig_height, fig_width, sub_i + 1, polar=True)
            )
            windrose(
                wind_dir,
                n_sectors,
                fig=subplots[-1],
                subfig=True,
                speed=speed,
                ticks=False,
                *args,
                **kwds
            )
        plt.title(str(time), fontsize="small")

    if normalize:
        # rescale all subfigs
        _yscale_subplots(fig, regrid=True)
        # remove all yticklabels (they are misplaced anyway)
        for subplot in subplots:
            former_labels = subplot.get_yticklabels()
            subplot.set_yticklabels([""] * len(former_labels))
    return fig


# -----------------------------------------------------------------------------#
def rel_ranks(values):
    """Returns ranks of values in the range [0,1]."""
    from scipy import stats

    return (stats.stats.rankdata(values) - 0.5) / len(values)


# -----------------------------------------------------------------------------#
def equal_num_bins(values, n_bins, round=True):
    """Returns bins so that values is cut into n_bins bins that each contain
    the same number of values' values."""
    bins = np.zeros(n_bins)
    bins[-1] = values.max()
    value_ranks = rel_ranks(values)
    for ii, perc in enumerate(np.linspace(0, 1, n_bins)[1:-1]):
        bins[ii + 1] = np.min(values[value_ranks > perc])
    return bins


# -----------------------------------------------------------------------------#
def scatter(wind_dirs, speed):
    fig = plt.figure()
    ax = plt.subplot(111, polar=True)
    wind_dirs = wind_dirs[np.isfinite(wind_dirs)]
    wind_dirs = -1 * wind_dirs + 450
    plt.scatter(
        wind_dirs,
        speed,
        marker="o",
        facecolors=(0, 0, 0, 0),
        edgecolors=(0, 0, 0, 0.5),
    )
    plt.ylim(0, speed.max())
    xicks = plt.xticks()[0]
    xlabels = list(range(90, -1, -45)) + list(range(315, 90, -45))
    plt.xticks(xicks, xlabels)
    yicks = plt.yticks()[0]
    plt.yticks(yicks, [str(ytick) for ytick in yicks])


# -----------------------------------------------------------------------------#
def windrose(
    wind_dirs,
    n_sectors=32,
    speed=None,
    speed_bins=6,
    fig=None,
    subfig=False,
    ticks=True,
    normalize=True,
    *args,
    **kwds
):
    """Plots a windrose (histogram with polar axes) of the given wind_dirs.
    If speed_bins is an int, equally filled speed_bins will be created."""
    fig = plt.figure() if fig is None else fig

    # transform directions to enable a the windrose representation
    # also works with "+ 90" but the resulting classes are less intuitive
    wind_dirs = -1 * wind_dirs + 450

    # enforce bins and ignore the classes given by np.histogram. we do not
    # want the sectors to change according to the data.
    bins = np.linspace(
        90 - 180.0 / n_sectors,
        450 + 180.0 / n_sectors,
        n_sectors + 2,
    )
    hist = np.histogram(wind_dirs, bins)[0]
    hist[0] += hist[-1]
    hist = hist[:-1]
    if normalize:
        hist = hist.astype(float) / np.sum(hist)
    if speed is not None:
        if type(speed_bins) == int:
            speed_bins = equal_num_bins(speed, speed_bins)
        s_hists = [
            np.histogram(
                speed[(wind_dirs > left) & (wind_dirs <= right)], speed_bins
            )[0]
            for left, right in zip(bins[:-1], bins[1:])
        ]
        s_hists[0] += s_hists[-1]
        s_hists = s_hists[:-1]
        # normalize all the classes to the frequency of wind in the individual
        # sectors
        # we do not need the speeds anymore! continue with the histogram
        speed = np.array(s_hists, dtype=float)
        # normalize by hand and not with density=True in np.histogram, because
        # "Note that the sum of the histogram values will not be equal to 1
        # unless bins of unity width are chosen; it is not a probability mass
        # function" (numpy documentation)
        speed /= speed.sum(axis=1)[:, np.newaxis]
        # make the speeds add up to the frequency of the wind falling in this
        # sector
        speed *= hist[:, np.newaxis]
    return _plot_dir_hist(
        hist, speed, speed_bins, fig, subfig, ticks, *args, **kwds
    )


# -----------------------------------------------------------------------------#
if __name__ == "__main__":
    # so lets do some examples..
    # fake up wind directions following a normal distribution
    dirs = np.random.normal(loc=180, scale=45, size=365 * 4)
    # overwrite values that are not in the range (0,360]
    dirs[(dirs < 0) & (dirs >= 360)] = 0
    # generate a datetime array (one year of 6h-timesteps) to plot seasonal
    # windroses
    datetimes = np.array(
        [
            datetime.datetime(2000, 1, 1) + datetime.timedelta(hours=hour)
            for hour in range(0, 365 * 4 * 6, 6)
        ]
    )

    # the most simple example
    windrose(dirs)
    # additional keyword arguments are passed to the plt.bar/ax.bar function.
    # here: no transparency for the bars
    windrose(dirs, alpha=1)
    # plots by month per default
    seasonal_windroses(datetimes, dirs, title="Wind directions by month")
    # plot by hour
    seasonal_windroses(
        datetimes, dirs, time_format_str="%H", title="Wind directions by hour"
    )

    # fake up some speed data
    weibull_ppf = lambda x, alpha, beta: (-np.log(1 - x) / alpha) ** (1 / beta)
    speeds = weibull_ppf(np.random.random(365 * 4), 1, 1.5)

    #    import wind_station as ws
    #    konst = ws.WindStation("Messstation Langenargen")
    #    seasonal_windroses(konst.datetime, konst.direction, time_format_str="%H",
    #                       speed=konst.speed)

    #    windrose(konst.direction)
    #    windrose(konst.direction, 32, konst.speed)
    #    speed_bins = np.linspace(0, konst.speed.max(), 5)
    ##    windrose(konst.direction[konst.speed < speed_bins[1]])
    #
    plt.show()



import numpy as np
import warnings


# ------------------------------------------------------------------------------#
def simulated_annealing(
    func,
    x0,
    args=(),
    k=0.99,
    M=None,
    step_length=None,
    T0=None,
    step_length0=5,
    feps=1e-5,
    constraints=None,
    callback=None,
):
    """Simulated Annealing algorithm with support for constraints, estimation
    for acceptance rate based initial temperature and gradient based
    calculation of the step lengths."""
    x0 = list(x0)
    n_dims = len(x0)
    M = 20 * n_dims if M is None else M
    if constraints:
        feasible = True
        for constraint in constraints:
            feasible *= constraint(x0)
        if not feasible:
            raise ValueError
    try:
        initial_obj = func(x0, *args)
    except ValueError:
        warnings.warn("Initial solution produced non-finite result.")
        raise

    # --------------------------------------------------------------------------#
    def get_new_candidate(solution):
        local_step_length = step_length

        def gen_new(old):
            return old + local_step_length * (np.random.random(n_dims) - 0.5)

        if constraints is None:
            return gen_new(solution[:])
        else:
            counter, counter_limit, meta_counter = 0, 100, 1
            # generate candidates until all constraints are met
            feasible = False
            while not feasible:
                new = gen_new(solution[:])
                feasible = True
                for constraint in constraints:
                    feasible *= constraint(new)
                counter += 1
                if counter > counter_limit:
                    print(
                        "Info: Could not find a feasible neighbor in %d "
                        % (counter_limit * 2**meta_counter)
                        + "iterations"
                    )
                    counter = 0
                    meta_counter += 1
                    counter_limit *= 2
                    # note, that this changes the step_length only in the
                    # local scope!
                    local_step_length *= 0.75
                if counter_limit * 2**meta_counter > 1e6:
                    warnings.warn("Giving up on finding a feasible neighbor")
                    raise RuntimeError
        return new

    # --------------------------------------------------------------------------#
    def initial_acceptance_rate(T0, best_obj_sofar=None, best_sol_sofar=None):
        """Return an acceptance rate for M timesteps starting at x0."""
        decisions = np.zeros(M, dtype=float)
        old_solution = x0[:]
        old_obj = func(old_solution, *args)
        counter, counter_limit, meta_counter = 0, 100, 1
        for ii in range(M):
            candidate = get_new_candidate(old_solution)
            counter += 1
            try:
                new_obj = func(candidate, *args)
            except ValueError:
                # do not accept solutions that produce non-finite output
                if counter > counter_limit:
                    warnings.warn(
                        "Warning: Got %d non-finite solutions "
                        % (counter_limit * 2**meta_counter)
                        + "during start-up."
                    )
                    counter = 0
                    meta_counter += 1
                    counter_limit *= 2
                continue
            if np.random.random() < np.exp((old_obj - new_obj) / T0):
                decisions[ii] = 1.0
                old_solution, old_obj = candidate, new_obj
                if best_obj_sofar is not None and (old_obj < best_obj_sofar):
                    best_obj_sofar, best_sol_sofar = old_obj, old_solution
        if best_obj_sofar is None:
            return decisions.mean()
        else:
            return decisions.mean(), best_obj_sofar, best_sol_sofar

    # --------------------------------------------------------------------------#
    def get_average_gradient(step_length0):
        """Average a gradient of the objective function by taking an M-step
        random walk around a solution."""
        if step_length0 is None:
            step_length0 = 1.0
        grad = np.zeros(n_dims, dtype=float)
        old_sol, old_obj = x0, initial_obj
        cur_step = 10
        for ii in range(M):
            dx = step_length0 * (np.random.random(n_dims) - 0.5)
            new_sol = old_sol + dx
            if constraints:
                feasible = True
                for constraint in constraints:
                    feasible *= constraint(new_sol)
                if not feasible:
                    step_length0 *= 0.75
                    ii -= 1
                    continue
            try:
                new_obj = func(new_sol, *args)
            except ValueError:
                cur_step *= 0.75
                continue
            grad += (new_obj - old_obj) / dx
            old_obj, old_sol = new_obj, new_sol
        grad /= np.sum(grad**2) ** 0.5
        return grad

    # --------------------------------------------------------------------------#
    def get_T0():
        """Estimates a starting temperature for which we get an acceptance rate
        that is  between 40-80%."""
        # take a random walk in the 'hood and check the objective values there
        hood, hood_obj = [get_new_candidate(x0)], [initial_obj]
        for ii in range(1, M):
            hood += [get_new_candidate(hood[ii - 1])]
            try:
                hood_obj += [func(hood[-1], *args)]
            except ValueError:
                ii -= 1
                continue
        hood_obj = np.array(hood_obj)
        # also, record the best solution and its value
        best_obj_sofar = np.min(hood_obj)
        best_sol_sofar = hood[np.argmin(hood_obj)]
        mean_exp_difference = np.mean(
            np.exp(initial_obj - hood_obj[hood_obj > initial_obj])
        )

        f_smaller = np.sum(hood_obj < initial_obj) / float(M)
        if f_smaller < 0.6:
            #            def rel_change(T0):
            #                return (f_smaller -
            #                        np.mean(np.exp((initial_obj - hood[hood > initial_obj])
            #                                       / T0)) -
            #                        .6) ** 2
            #            T0 = optimize.fmin(rel_change, 1)[0]
            T0 = np.log(0.6 - f_smaller) / np.log(mean_exp_difference)
        else:
            T0 = np.log(f_smaller) / np.log(mean_exp_difference)

        if T0 <= 0:
            warnings.warn("T0 was %.3f. Starting over" % T0)
            return get_T0()

        find_temp_counter = 0
        accept, best_obj_sofar, best_sol_sofar = initial_acceptance_rate(
            T0, best_obj_sofar, best_sol_sofar
        )
        while (accept < 0.4) or (accept > 0.8):
            if accept < 0.4:
                T0 *= 1 + 2 * (1 - accept)
            else:
                T0 /= 1 + 2 * (1 - accept)
            accept, best_obj_sofar, best_sol_sofar = initial_acceptance_rate(
                T0, best_obj_sofar, best_sol_sofar
            )
            find_temp_counter += 1
        print(
            "Found suitable starting temperature after %d objective "
            % (M * (find_temp_counter + 1) + M)
            + "function evaluations."
        )
        return T0, best_obj_sofar, best_sol_sofar

    # --------------------------------------------------------------------------#
    if step_length is None:
        step_length = get_average_gradient(step_length0)
        if step_length0 is not None:
            step_length *= step_length0
        print(
            "Step lengths are: "
            + ", ".join("%.3f" % abs(step) for step in step_length)
        )
    if T0 is None:
        T0, best_obj_sofar, best_sol_sofar = get_T0()
    else:
        best_obj_sofar, best_sol_sofar = initial_obj, x0

    print("Initial obj. value: %.3f" % initial_obj)
    print("Best obj. value during start-up: %.3f" % best_obj_sofar)
    old_obj = conv_obj = initial_obj
    conv_sol = best_sol_sofar
    rel_change, old_sol, T, m = feps + 1, best_sol_sofar, T0, 0
    iteration_counter = 0
    counter, counter_limit, meta_counter = 0, 100, 1
    while (iteration_counter < 5 * M) or (abs(rel_change) > feps):
        try:
            new_sol = get_new_candidate(old_sol)
        except RuntimeError:
            break
        counter += 1
        try:
            new_obj = func(new_sol, *args)
        except ValueError:
            if counter > counter_limit:
                warnings.warn(
                    "Warning: Discarded %d solutions that produced "
                    % (counter_limit * 2**meta_counter)
                    + "a non-finite objective function value."
                )
                counter = 0
                meta_counter += 1
                counter_limit *= 2
            continue

        if (new_obj < old_obj) or np.random.random() < np.exp(
            (old_obj - new_obj) / T
        ):
            old_sol, old_obj = new_sol, new_obj
            counter_limit, meta_counter = 100, 1
            if new_obj < best_obj_sofar:
                best_sol_sofar, best_obj_sofar = new_sol, new_obj
        m += 1

        if m == M:
            m = 0
            T *= k
            if np.all(conv_sol != old_sol):
                step_length = (conv_obj - old_obj) / (conv_sol - old_sol)
                # include a bit of noise, to avoid zeros in step_length
                step_length += (
                    0.5 * step_length.mean() * (np.random.random(n_dims) - 0.5)
                )
                step_length /= np.sum(step_length**2) ** 0.5
                step_length0 *= k
                step_length *= step_length0
            rel_change = (old_obj - conv_obj) / abs(conv_obj)
            conv_obj, conv_sol = old_obj, old_sol
            if callback:
                callback(conv_obj)
            print(
                "Iteration: %d, obj. value: %.3f, Temp: %f, "
                % (iteration_counter + 1, old_obj, T)
                + "rel. change: %.3f, " % rel_change
            )
        #                     +
        #                   "Step lengths: %s" %
        #                   ", ".join("%.3f" % abs(sl) for sl in step_length)

        iteration_counter += 1
    print(
        "Converged after %d iterations. Obj. value: %.3f (Initial "
        % (iteration_counter + 1, best_obj_sofar)
        + "obj.: %.3f)" % initial_obj
    )
    print("Solution: ", best_sol_sofar)
    return best_sol_sofar


# ------------------------------------------------------------------------------#



from cpython.datetime cimport import_datetime
import_datetime()

import numpy as np
cimport numpy as np
cimport cython
from cython.parallel import prange
from cpython cimport datetime
from datetime import datetime as pydatetime

cdef double min_per_day = 24. * 60
cdef double sec_per_day = min_per_day * 60
cdef double msec_per_day = sec_per_day * 1e6


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double[::1] datetime2doy(list dt):
    """ Extracts the day of year as a float from the given datetimes.
    """
    cdef unsigned int i, N = len(dt)
    cdef double[::1] doys = np.empty(N, dtype=float)
    cdef datetime.datetime sub_dt
    for i in range(N):
        sub_dt = dt[i]
        tt = sub_dt.timetuple()
        doys[i] = (<int>tt.tm_yday +
                   <int>tt.tm_hour / 24. +
                   <int>tt.tm_min / min_per_day +
                   <int>tt.tm_sec / sec_per_day +
                   <int>sub_dt.microsecond / msec_per_day)
    return doys


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef doy_distance(double doy0, double[::1] doys):
    cdef int t, T = len(doys)
    cdef double dist
    dists = np.empty(T, dtype=np.double)
    cdef double[::1] dist_view = dists
    for t in prange(T, nogil=True, schedule="guided"):
        dist = (doy0 - doys[t]) % 365.
        if dist > (365. / 2):
            dist_view[t] = 365 - dist
        else:
            dist_view[t] = dist
    return dists
    
# def periodic_distance(x1, x2, period):
#     """

#     Examples
#     --------
#     >>> periodic_distance(0, 23, 24)
#     array(1.0)
#     """
#     period = float(period)
#     dist = (x1 - x2) % period
#     return np.where(dist > period / 2, period - dist, dist)
